"""

Author: Martin Benes, Benedikt Lorch
Affiliation: University of Innsbruck
"""

import numpy as np
import scipy.fftpack

from .common import entropy


def nzAC(dct: np.ndarray) -> int:
    """Computes number of non-zero DCT AC coefficients from 4D DCT tensor.

    :param dct: array of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :type dct: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :return: number of non-zero DCT AC coefficients
    :rtype: int

    :Example:

    >>> cl.tools.nzAC(jpeg0.Y)
    """
    assert len(dct.shape) == 4, "Expected 4D DCT input array"
    return (dct != 0).sum() - (dct[:, :, 0, 0] != 0).sum()


def AC(dct: np.ndarray) -> int:
    """Computes number of DCT AC coefficients from 4D DCT tensor.

    :param dct: array of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :type dct: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :return: number of DCT AC coefficients
    :rtype: int

    :Example:

    >>> cl.tools.AC(jpeg0.Y)
    """
    assert len(dct.shape) == 4, "Expected 4D DCT input array"
    return dct.size - np.prod(dct.shape[:2])


def embedding_rate(
    dct_c: np.ndarray,
    dct_s: np.ndarray,
    q: int = 3,
) -> float:
    """Estimate embedding rate in DCT domain.

    :param dct_c:
    :type dct_c: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param dct_s:
    :type dct_s: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param q:
    :type q: int
    :return:
    :rtype: float

    :Example:

    >>> # TODO
    """
    # no changes
    if (dct_c == dct_s).all():
        return 0.
    # embeddable elements
    embeddable = nzAC(dct_c)
    # binary bound
    if q == 2:
        # number of changes
        changes = (dct_c != dct_s).sum()
        # change rate
        betas = [changes/embeddable]

    # ternary bound
    elif q == 3:
        # number of changes
        changesP1 = (dct_c < dct_s).sum()
        changesM1 = (dct_c > dct_s).sum()
        # change rate
        betas = [
            changesP1 / embeddable,
            changesM1 / embeddable,
        ]

    else:
        raise NotImplementedError(f'{q}-ary code not implemented')

    # embedding rate
    alpha = entropy(*betas)
    return alpha


def Lambda(x: float) -> np.array:
    """DCT scaling function.

    :param x: Input value.
    :type x: float
    :return:
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:
    >>> Lambda(0) # -> 0.70710678
    >>> Lambda(-1) # -> 1.
    """
    return 1. if x != 0 else 1/np.sqrt(2)


def dct2(a: np.ndarray) -> np.ndarray:
    """"""
    a = scipy.fftpack.dct(a, axis=0, norm='ortho')
    a = scipy.fftpack.dct(a, axis=1, norm='ortho')
    return a


def idct2(a: np.ndarray) -> np.ndarray:
    """"""
    a = scipy.fftpack.idct(a, axis=0, norm='ortho')
    a = scipy.fftpack.idct(a, axis=1, norm='ortho')
    return a


def jpeglib_to_jpegio(dct_coeffs: np.ndarray) -> np.ndarray:
    """
    Convert DCT coefficient array shape from 4D jpeglib format to 2D jpegio format
    :param dct_coeffs: DCT coefficients of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :return: DCT coefficients reshaped to [num_vertical_blocks * 8, num_horizontal_blocks * 8]

    :Example:

    >>> # TODO
    """
    num_vertical_blocks, num_horizontal_blocks, block_height, block_width = dct_coeffs.shape
    assert block_height == 8, "Expected block height of 8"
    assert block_width == 8, "Expected block width of 8"

    # Transpose from [num_vertical_blocks, num_horizontal_blocks, 8, 8] to [num_vertical_blocks, 8, num_horizontal_blocks, 8]
    dct_coeffs = dct_coeffs.transpose((0, 2, 1, 3))

    # Reshape to [num_vertical_blocks * 8, num_horizontal_blocks * 8]
    dct_coeffs = dct_coeffs.reshape((num_vertical_blocks * 8, num_horizontal_blocks * 8))

    return dct_coeffs


def jpegio_to_jpeglib(dct_coeffs: np.ndarray) -> np.ndarray:
    """
    Convert DCT coefficient array shape from 2D jpegio format to 4D jpeglib format
    :param dct_coeffs: DCT coefficients of shape [num_vertical_blocks * 8, num_horizontal_blocks * 8]
    :return: DCT coefficients reshaped to [num_vertical_blocks, num_horizontal_blocks, 8, 8]

    :Example:

    >>> # TODO
    """
    # From jpegio 2D to jpeglib 4D
    assert dct_coeffs.shape[0] % 8 == 0
    assert dct_coeffs.shape[1] % 8 == 0

    num_vertical_blocks = dct_coeffs.shape[0] // 8
    num_horizontal_blocks = dct_coeffs.shape[1] // 8

    # Reshape from [num_vertical_blocks * 8, num_horizontal_blocks * 8] to [num_vertical_blocks, 8, num_horizontal_blocks, 8]
    dct_coeffs = dct_coeffs.reshape((num_vertical_blocks, 8, num_horizontal_blocks, 8))

    # Reorer to [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    dct_coeffs = dct_coeffs.transpose((0, 2, 1, 3))

    return dct_coeffs


def compute_DCT_mat() -> np.ndarray:
    """
    Computes the 8x8 DCT matrix

    :return: ndarray of shape [8, 8]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> # TODO
    """
    [col, row] = np.meshgrid(range(8), range(8))
    dct_mat = 0.5 * np.cos(np.pi * (2 * col + 1) * row / (2 * 8))
    dct_mat[0, :] = dct_mat[0, :] / np.sqrt(2)
    return dct_mat


def block_dct2(
    spatial_blocks: np.ndarray,
    dct_mat: np.ndarray = None,
) -> np.ndarray:
    """
    Apply 2D DCT to image blocks
    :param spatial_blocks: ndarray of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :param dct_mat: ndarray of shape [8, 8]. If None, the DCT matrix is computed on-the-fly.
    :return: DCT coefficients of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]

    :Example:

    >>> # TODO
    """
    if dct_mat is None:
        dct_mat = compute_DCT_mat()

    DCT_mat_left = dct_mat[None, None, :, :]
    DCT_mat_right = (dct_mat.T)[None, None, :, :]

    dct_coeffs = DCT_mat_left @ spatial_blocks @ DCT_mat_right

    return dct_coeffs


def block_idct2(
    dct_coeffs: np.ndarray,
    dct_mat: np.ndarray = None,
) -> np.ndarray:
    """Apply 2D inverse DCT to image blocks

    :param dct_coeffs: ndarray of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :type dct_coeffs: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param dct_mat: ndarray of shape [8, 8]. If None, the DCT matrix is computed on-the-fly.
    :type dct_map: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :return: spatial blocks of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> # TODO
    """
    if dct_mat is None:
        dct_mat = compute_DCT_mat()

    DCT_mat_left = (dct_mat.T)[None, None, :, :]
    DCT_mat_right = dct_mat[None, None, :, :]

    spatial_blocks = DCT_mat_left @ dct_coeffs @ DCT_mat_right

    return spatial_blocks



"""
Implementation of the HUGO steganography method as described in

T. Pevny, T. Filler, P. Bas
"Using High-Dimensional Image Models to Perform Highly Undetectable Steganography"
ACM Information Hiding 2010
https://hal.science/hal-00541353/document

Author: Benedikt Lorch, Martin Benes
Affiliation: University of Innsbruck

This implementation builds on the original Matlab implementation provided by the paper authors. Please find that license of the original implementation below.
-------------------------------------------------------------------------
Copyright (c) 2013 DDE Lab, Binghamton University, NY. All Rights Reserved.
Permission to use, copy, modify, and distribute this software for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that this copyright notice appears in all copies. The program is supplied "as is," without any accompanying services from DDE Lab. DDE Lab does not warrant the operation of the program will be uninterrupted or error-free. The end-user understands that the program was developed for research purposes and is advised not to rely exclusively on the program for any reason. In no event shall Binghamton University or DDE Lab be liable to any party for direct, indirect, special, incidental, or consequential damages, including lost profits, arising out of the use of this software. DDE Lab disclaims any warranties, and has no obligations to provide maintenance, support, updates, enhancements or modifications.
-------------------------------------------------------------------------
"""  # noqa: E501

import numpy as np

from . import _costmap
from ..simulate import _ternary


def simulate_single_channel(
    x0: np.ndarray,
    alpha: float,
    *,
    sigma: float = 1,
    gamma: float = 1,
    wet_cost: float = 1e8,
    **kw,
) -> np.ndarray:
    """Simulates HUGO embedding into a single channel.

    HUGO was introduced in
    T. Pevny, et al. Using High-Dimensional Image Models to Perform Highly Undetectable Steganography.
    ACM IH, 2010.

    The details of the methods are described in the
    `glossary <https://conseal.readthedocs.io/en/latest/glossary.html#highly-undetectable-stego>`__.

    :param x0: uncompressed (pixel) cover image
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param alpha: embedding rate
        in bits per pixel
    :type alpha: float
    :param wet_cost: wet cost for unembeddable coefficients
    :type wet_cost: float
    :param kw: remaining keyword parameters are passed to simulator
    :type kw: dict
    :return: modified stego image
        of shape [height, width]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> x1 = cl.hugo.simulate_single_channel(
    ...     x0=x0,  # pixels
    ...     alpha=0.4,  # alpha
    ...     seed=12345)  # seed
    """
    #
    if alpha == 0:
        return x0

    # Compute cost
    rhos = _costmap.compute_cost_adjusted(
        x0,
        sigma=sigma,
        gamma=gamma,
        wet_cost=wet_cost,
    )

    # Simulate
    delta = _ternary.ternary(
        rhos=rhos,
        alpha=alpha,
        n=x0.size,
        **kw,
    )
    return x0 + delta.astype('uint8')



"""

Author: Benedikt Lorch, Martin Benes
Affiliation: University of Innsbruck
"""

import numpy as np
import scipy.signal
# from scipy.signal import correlate2d
from . import _defs
from .. import tools
# from stegolab2.jmipod.costmap import wiener2, estimate_variance


def estimate_variance(
    image: np.ndarray,
    block_size: int,
    degree: int,
) -> np.ndarray:
    """
    Estimate the pixels' variance by fitting a local parametric model based on a 2D-DCT (trigonometric polynomial) to the neighbors of each residual value.
    :param image: image residual
    :type image: np.ndarray
    :param block_size: size of the local neighborhood, denoted as p in the paper. See Fig. 2.
        Small p: extreme content adaptivity;
        Medium p: Medium content adaptivity;
        Large p: Low content adaptivity.
    :type block_size: int
    :param degree: degree of the polynomial
    :type degree: int
    :return: estimated variance per pixel
    :rtype:
    """
    assert block_size % 2 != 0, 'block dimensions must be odd'
    assert degree <= block_size, 'number of basis vectors exceeds block size'

    # Number of parameters per p x p block, see sentence below Eq. 26
    q = degree * (degree + 1) // 2

    base_mat = np.zeros((block_size, block_size))
    base_mat[0, 0] = 1

    # Construct G matrix of size [p * p, q]. G defines the parametric model with 2-D discrete trigonometrical polynomial functions. See Eq. (25) and Eq. (26).
    G = np.zeros((block_size ** 2, q))
    k = 0
    for xshift in range(degree):
        for yshift in range(degree - xshift):
            base_mat_rolled = np.roll(base_mat, shift=(xshift, yshift), axis=(0, 1))
            G[:, k] = tools.dct.idct2(base_mat_rolled).flatten(order="F")
            k += 1

    # Symmetry-pad image by block_size / 2 in all directions
    pad_size = block_size // 2
    img_padded = np.pad(image, (pad_size, pad_size), "symmetric")

    # Solve G.T G x = G.T for x.
    # x = (G.T G)^{-1} G.T
    x = np.linalg.solve(G.T @ G, G.T)

    # P_G_orth = I - G x = I - G (G^T G)^{-1} G.T
    # P_G_orth represents the orthogonal projection ontho the p^2 - q dimensional subspace spanned by the left null space of G.
    P_G_orth = np.eye(block_size ** 2) - (G @ x)

    sliding_window_views = _defs.im2col(
        img_padded,
        window_shape=(block_size, block_size),
    )

    # Eq. 24
    estimated_variance = np.sum((P_G_orth @ sliding_window_views) ** 2, axis=0) / (block_size ** 2 - q)
    estimated_variance = estimated_variance.reshape(image.shape)

    # The estimated value of the variance is attributed to the center pixel.
    return estimated_variance


def compute_cost(x0: np.ndarray) -> np.ndarray:
    """Produces MiPOD cost. Not a cost like HILL, but a Fisher information.

    :param x0: uncompressed (pixel) cover image
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :return: cost for +-1 change
        of shape [height, width]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> fi = cl.mipod.compute_cost(x0=x0)
    """
    x0 = x0.astype(float)

    # Compute wiener residual
    # Suppressed the content, the residual is zero mean
    # The residual still contains remnants of the content around edges and in textured areas.
    wiener_residual = x0 - _defs.wiener2(x0, kernel_size=(2, 2))

    # Estimate variance: Fit a local parameteric model to the neighbors of each pixel value
    variance = estimate_variance(wiener_residual, block_size=9, degree=9)

    # Set minimum variance, Eq. 27
    variance = np.clip(variance, a_min=0.01, a_max=None)

    # Compute Fisher information (I = 2 / sigma_n ** 4).
    fisher_information = 1 / variance ** 2

    # average_kernel =
    fisher_information = scipy.signal.correlate2d(
        fisher_information,
        np.ones((7, 7)) / 7 ** 2,
        mode="same", boundary="symm",
    )

    return fisher_information



#!/usr/bin/env python3
"""
Implementation of WOW steganographic embedding.

Code inspired by Holub's Matlab implementation: https://dde.binghamton.edu/download/stego_algorithms/

Designing Steganographic Distortion Using Directional Filters.
V. Holub, and J. Fridrich. http://www.ws.binghamton.edu/fridrich/research/WOW_rewritten_ver_WIFS_02.pdf
"""

import numpy as np
import scipy.signal
from typing import Tuple

from .. import tools


def compute_cost(
    x0: np.ndarray,
    *,
    p: float = -1,
) -> Tuple[np.ndarray]:
    """Computes WOW cost.

    :param x0: uncompressed (pixel) cover image,
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param p: parameter for reciprocal Hoelder norm
    :type p: float
    :return: cost for +-1 change,
        of shape [height, width]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> # TODO
    """
    assert len(x0.shape) == 2, 'single channel expected'

    # 2D wavelet filters (Daubechies 8)
    _, F = tools.spatial.daubechies8()

    # add padding
    padSize = np.max([f.shape for f in F])
    x0_padded = np.pad(x0, padSize, 'symmetric')

    # compute directional residual and suitability \xi for each filter
    xi = []
    for fIndex in range(3):

        # compute residual
        R = scipy.signal.convolve2d(
            x0_padded,
            F[fIndex],
            mode='same', boundary='symm'
        )

        # compute sustability
        xi.append(
            scipy.signal.convolve2d(
                np.abs(R),
                np.rot90(np.abs(F[fIndex]), k=2),
                mode='same', boundary='symm'
            )
        )
        # correct the suitability shift if filter size is even
        if F[fIndex].shape[0] % 2 != 0:
            xi[fIndex] = np.roll(xi[fIndex], 1, axis=0)
        if F[fIndex].shape[1] % 2 != 0:
            xi[fIndex] = np.roll(xi[fIndex], 1, axis=1)
        # remove padding
        x0_center = [
            (xi[fIndex].shape[i]-x0.shape[i]) >> 1
            for i in range(2)
        ]
        xi[fIndex] = xi[fIndex][
            x0_center[0]+1:-x0_center[0]+1,
            x0_center[1]+1:-x0_center[1]+1,
        ]

    # compute embedding costs \rho
    rho = np.sum([xi[i]**p for i in range(3)], axis=0)**(-1/p)
    # rho = np.expand_dims(rho, 2)
    return rho


def compute_cost_adjusted(
    x0: np.ndarray,
    wet_cost: float = 1e10,
    **kw,
) -> Tuple[np.ndarray]:
    """Computes WOW cost with wet-cost adjustments.

    :param x0: uncompressed (pixel) cover image,
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param wet_cost: wet cost for unembeddable coefficients
    :type wet_cost: float
    :return: costs for +1 and -1 changes,
        of shape [height, width]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> rhos = cl.wow.compute_cost_adjusted(x0=x0)
    """
    assert len(x0.shape) == 2, 'single channel expected'

    # process input
    x0 = x0.astype('float32')

    # Compute costmap
    rho = compute_cost(x0=x0, **kw)

    # Assign wet cost
    rho[np.isinf(rho) | np.isnan(rho) | (rho > wet_cost)] = wet_cost

    # Do not embed +1 if the pixel has max value
    rho_p1 = np.copy(rho)
    rho_p1[x0 >= 255] = wet_cost

    # Do not embed -1 if the pixel has min value
    rho_m1 = np.copy(rho)
    rho_m1[x0 <= 0] = wet_cost

    # return costs
    return rho_p1, rho_m1



"""
Implementation of the nsF5 steganography method as described in

J. Fridrich, T. Pevny, and J. Kodovsky.
"Statistically undetectable JPEG steganography: Dead ends, challenges, and opportunities"
Multimedia & Security, 2007
http://dde.binghamton.edu/kodovsky/pdf/Fri07-ACM.pdf

Author: Benedikt Lorch, Martin Benes
Affiliation: University of Innsbruck
"""  # noqa: E501

import numpy as np
from typing import Tuple

from .. import tools


def probability(
    y0: np.ndarray,
    alpha: float,
) -> np.ndarray:
    """Compute the nsF5 probability map.

    :param y0: quantized cover DCT coefficients,
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :type y0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param alpha: embedding rate,
        in bits per nzAC coefficient
    :type alpha: float
    :return: tuple ((p_p1, p_m1), None), where
        p_p1 is the probability of +1 change
        p_m1 is the probability of -1 change.
    :rtype: tuple of `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> # TODO
    """
    assert len(y0.shape) == 4, 'DCT must have 4 dimensions'
    assert y0.shape[2] == y0.shape[3] == 8, 'expected blocks of size 8x8'

    # No embedding
    if np.isclose(alpha, 0):
        return np.zeros_like(y0)

    # Compute change rate on bound
    beta = tools.inv_entropy(alpha)

    # Number of nonzero AC DCT coefficients
    nzAC = tools.dct.nzAC(y0)
    if nzAC == 0:
        raise ValueError('There are no non-zero AC coefficients for embedding')

    # probability map
    p = np.ones(y0.shape, dtype='float64') * beta

    # do not change zeros or DC mode
    p[y0 == 0] = 0
    p[:, :, 0, 0] = 0

    # substract absolute value
    p_p1, p_m1 = p.copy(), p.copy()
    p_p1[y0 > 0] = 0
    p_m1[y0 < 0] = 0

    return (p_p1, p_m1), None


def compute_cost(
    y0: np.ndarray,
    *,
    wet_cost: float = 10**10
) -> np.ndarray:
    """Compute the nsF5 cost.

    :param y0: quantized cover DCT coefficients,
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :type y0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param wet_cost: wet cost for unembeddable coefficients
    :type wet_cost: float
    :return: probability of +-1 change,
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :rtype: tuple of `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> # TODO
    """
    assert len(y0.shape) == 4, 'DCT must have 4 dimensions'
    assert y0.shape[2] == y0.shape[3] == 8, 'expected blocks of size 8x8'

    # cost
    rho = np.ones(y0.shape, dtype='float64')

    # do not change zeros or DC mode
    rho[y0 == 0] = wet_cost
    rho[:, :, 0, 0] = wet_cost

    return rho


def compute_cost_adjusted(
    y0: np.ndarray,
    *,
    wet_cost: float = 10**10,
) -> Tuple[np.ndarray]:
    """Compute nsF5 cost with wet-cost adjustments.

    :param y0: quantized cover DCT coefficients,
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :type y0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param wet_cost: wet cost for unembeddable coefficients
    :type wet_cost: float
    :return: probability of +-1 change,
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :rtype: tuple of `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> rhos = cl.nsF5.compute_cost_adjusted(jpeg0.Y)
    >>> jpeg1.Y = jpeg0.Y + cl.simulate.ternary(rhos=rhos, alpha=.4, seed=12345)
    """
    # Compute costmap
    rho = compute_cost(y0=y0, wet_cost=wet_cost)

    # Assign wet cost
    rho[np.isinf(rho) | np.isnan(rho) | (rho > wet_cost)] = wet_cost

    # Subtract absolute value
    rho_p1, rho_m1 = rho.copy(), rho.copy()
    rho_p1[y0 > 0] = wet_cost
    rho_m1[y0 < 0] = wet_cost

    return rho_p1, rho_m1



"""High-level interface of simulated HILL embedding.

Author: Martin Benes, Benedikt Lorch
Affiliation: University of Innsbruck
"""

import numpy as np

from . import _costmap
from ..simulate import _ternary


def simulate_single_channel(
    x0: np.ndarray,
    alpha: float,
    *,
    wet_cost: float = 10**10,
    **kw,
) -> np.ndarray:
    """Simulates HILL embedding into a single channel.

    HILL was introduced in
    B. Li, et al. A New Cost Function For Spatial Image Steganography.
    IEEE ICIP, 2014.

    The details of the methods are described in the
    `glossary <https://conseal.readthedocs.io/en/latest/glossary.html#high-low-low>`__.

    :param x0: uncompressed (pixel) cover image
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param alpha: embedding rate
        in bits per pixel
    :type alpha: float
    :param wet_cost: wet cost for unembeddable coefficients
    :type wet_cost: float
    :return: stego pixels
        of shape [height, width, channels]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> x1 = cl.hill.simulate_single_channel(
    ...     x0=x0,  # cover pixels
    ...     alpha=.4,  # embedding rate [bpp]
    ...     seed=12345)  # seed for PSNR
    """
    if alpha == 0:
        return x0

    # Compute cost
    rhos = _costmap.compute_cost_adjusted(
        x0=x0,
        wet_cost=wet_cost,
    )

    # Simulate
    ps, _ = _ternary.probability(
        rhos=rhos,
        alpha=alpha,
        n=x0.size,
    )
    delta = _ternary.simulate(ps=ps, **kw)
    return x0 + delta.astype('uint8')



"""Implementation of matrix embedding.

Author: Martin Benes
Affiliation: University of Innsbruck
"""

import scipy.optimize


def efficiency(
    alpha: float
) -> float:
    """Approximates embedding efficiency for Hamming code.

    Assumes last block to be padded.

    :param alpha: embedding rate in bits per element
    :type alpha: float
    :return: embedding efficiency in bits per change
    :rtype: float

    :Example:

    >>> e = cl.coding.efficiency(0.4)  # e at alpha=0.4
    """
    assert alpha >= 1e-3, 'instable for very low embedding rates'

    def mse_p(p) -> float:
        """SS error from blocksize."""
        return (p / (2**p - 1) - alpha)**2

    p = scipy.optimize.fminbound(mse_p, 1, 100, xtol=1e-4)
    return p / (1 - 2.**-p)



import numpy as np
from .dct import block_idct2


def decompress_channel(y, qt=None):
    """Decompress a single channel.

    :param y: quantized DCT coefficients,
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :param qt: quantization table,
        of shape [8, 8]
    :return: decompressed channel,
        of shape [num_vertical_blocks * 8, num_horizontal_blocks * 8].
        The values should be in the range [0, 255], although some values may exceed this range.
    """

    # Validate quantization table
    if qt is not None:
        assert (
                len(qt.shape) == 2
                and qt.shape[0] == 8
                and qt.shape[1] == 8), "Expected quantization table of shape [8, 8]"

    num_vertical_blocks, num_horizontal_blocks = y.shape[:2]

    # If quantization table is given, dequantize the DCT coefficients
    if qt is not None:
        y = y * qt[None, None, :, :]

    # Inverse DCT
    x = block_idct2(y)

    # Reorder blocks to obtain image
    x = np.transpose(x, axes=[0, 2, 1, 3]).reshape(num_vertical_blocks * 8, num_horizontal_blocks * 8)

    # Level shift
    x += 128

    return x



"""
Implementation of the UERD steganography method as described in

L. Guo, J. Ni, W. Su, C. Tang and Y.-Q. Shi
"Using Statistical Image Model for JPEG Steganography: Uniform Embedding Revisited"
IEEE Transactions on Information Forensics and Security, 2015
https://ieeexplore.ieee.org/document/7225122

The idea of uniform embedding distortion (UED) is to spread the changes across all the DCT coefficients (including the DC, zero and non-zero AC coefficients).

The uniform embedding revisited distortion (UERD) incorporates the complexity of the specific DCT block and the specific DCT mode.

Author: Martin Benes, Benedikt Lorch
Affiliation: University of Innsbruck
"""  # noqa: E501

import numpy as np
from scipy.signal import convolve2d
import typing

from .. import tools


def compute_block_energies(
    y: np.ndarray,
    qt: np.ndarray,
) -> np.ndarray:
    """Compute block energy as described in Eq. 3

    :param y: quantized DCT coefficients,
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :type y: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param qt: quantization table,
        of shape [8, 8]
    :type qt: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :return: block energies,
        of shape [num_vertical_blocks, num_horizontal_blocks]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> rho = cl.uerd._costmap.compute_block_energies(
    ...   y0=jpeg0.Y,  # DCT
    ...   qt=jpeg0.qt[0])  # QT
    """

    num_vertical_blocks, num_horizontal_blocks = y.shape[:2]

    # Dequantize DCT coefficients
    y_deq = y * qt[None, None, :, :]

    # Flatten last dimension
    y_deq = y_deq.reshape(num_vertical_blocks, num_horizontal_blocks, -1)

    # Calculate block energies, excluding the DC coefficient
    block_energies = np.sum(np.abs(y_deq)[:, :, 1:], axis=2)

    return block_energies


def compute_cost(
    y0: np.ndarray,
    qt: np.ndarray,
) -> np.ndarray:
    """Compute the UERD cost.

    Check Eq. 4 of the paper.

    :param y0: quantized cover DCT coefficients,
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :type y0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param qt: quantization table,
        of shape [8, 8]
    :type qt: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :return: embedding cost,
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> # TODO
    """
    # Compute block energies
    block_energies = compute_block_energies(y0, qt)
    num_vertical_blocks, num_horizontal_blocks = block_energies.shape

    # Compute energy in 8-neighborhood using a convolution
    neighborhood = np.array([[1, 1, 1], [1, 0, 1], [1, 1, 1]])

    block_energies_in_neighborhood = convolve2d(block_energies, neighborhood, mode="same", boundary="symm")

    # Block rho of shape [num_vertical_blocks, num_horizontal_blocks]
    block_rho = block_energies + 0.25 * block_energies_in_neighborhood

    # Mode rho of shape [64]
    mode_rho = np.zeros(64, dtype=float)
    # DC
    mode_rho[0] = 0.5 * (qt[0, 1] + qt[1, 0])
    # AC
    mode_rho[1:] = qt.flatten()[1:]

    # Combine block and mode rho
    # Division by zero should result in inf cost
    rho = np.divide(
        mode_rho[None, None, :],
        block_rho[:, :, None],
        out=np.full(shape=(num_vertical_blocks, num_horizontal_blocks, 64), fill_value=np.inf, dtype=float),
        where=block_rho[:, :, None] > 0
    )

    rho = rho.reshape(num_vertical_blocks, num_horizontal_blocks, 8, 8)

    return rho


def compute_cost_adjusted(
    y0: np.ndarray,
    qt: np.ndarray,
    *,
    wet_cost: float = 10**13,
    avoid_saturated: bool = False,
) -> typing.Tuple[np.ndarray, np.ndarray]:
    """Compute the adjusted J-UNIWARD cost for ternary embedding.

    :param y0: quantized cover DCT coefficients,
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :type y0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param qt: quantization table,
        of shape [8, 8]
    :type qt: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param wet_cost: cost for unembeddable coefficients
    :type wet_cost: float
    :param avoid_saturated: hard-sets blocks with saturated pixels to wet
    :type avoid_saturated: bool
    :return: embedding costs of +1 and -1 changes,
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :rtype: tuple of `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> rho_p1, rho_m1 = cl.uerd.compute_cost_adjusted(
    ...   y0=jpeg0.Y,      # DCT
    ...   qt=jpeg0.qt[0])  # QT
    """
    # Compute embedding cost rho
    rho = compute_cost(y0, qt)

    # Adjust embedding costs
    rho[np.isinf(rho) | np.isnan(rho) | (rho > wet_cost)] = wet_cost
    if avoid_saturated:
        x0 = tools.decompress_channel(y0)  # decompress DCT
        rho[(tools.jpegio_to_jpeglib(x0) == 0).any(axis=(2, 3))] = wet_cost
        rho[(tools.jpegio_to_jpeglib(x0) == 255).any(axis=(2, 3))] = wet_cost

    # Do not embed +1 if the DCT coefficient has maximum value
    rho_p1 = np.copy(rho)
    rho_p1[y0 >= 1023] = wet_cost

    # Do not embed -1 if the DCT coefficient has minimum value
    rho_m1 = np.copy(rho)
    rho_m1[y0 <= -1023] = wet_cost

    return rho_p1, rho_m1



"""

Author: Martin Benes
Affiliation: University of Innsbruck
"""

from . import hamming
from . import wpc
from . import trellis

from .. import tools


def efficiency(alpha: float) -> float:
    """Calculates the embedding efficiency at the bound.

    :param alpha: embedding rate, at bits per element
    :type alpha: float
    :return: embedding efficiency, at bits per element
    :rtype: float

    :Example:

    >>> e = cl.coding.efficiency(0.4)  # e at alpha=0.4
    """
    return alpha / tools.inv_entropy(alpha)



""""""

import hashlib
import numpy as np
import typing


def blocks(
    y: np.ndarray,
    key: int = None,
) -> typing.Tuple[int]:
    """Returns indices of permuted coefficients."""

    # permute
    rng = np.random.default_rng(key)
    perm = rng.permutation(y.size)

    # dimensions
    Hb, Wb, _, _ = y.shape
    wb = perm % 8
    hb = perm // 8 % 8
    w = perm // 64 % Wb
    h = perm // 64 // Wb
    return h, w, hb, wb


def iterate_ac(
    dct: np.ndarray,
    key: int,
    skipzero: bool = True
):
    """Generator for iterating ACs/nzACs."""
    # permutative straddling
    perm = blocks(dct, key)

    # iterate coefficients
    for cover_it in range(dct.size):

        # get cover element
        h, w, hb, wb = [pj[cover_it] for pj in perm]

        # skip
        if hb == 0 and wb == 0:  # DC
            continue
        if skipzero and dct[h, w, hb, wb] == 0:  # zero
            continue

        # append to window
        yield h, w, hb, wb


def iterate(
    x: np.ndarray,
    key: int = None,
):
    # flip
    flip = len(x.shape) == 4

    # shuffle
    perm = np.linspace(0, x.size-1, x.size, dtype='int64')
    if key:
        rng = np.random.default_rng(key)
        perm = rng.permutation(x.size)

    # prepare shapes
    mod = np.array(x.shape)
    if flip:
        mod = np.flip(mod)
    denom = np.hstack([[1], np.cumprod(mod)[:-1]])

    # iterate elements
    for it in perm:

        # get index
        idx = tuple([
            it // denom[i] % mod[i]
            for i in range(len(x.shape))
        ])

        if flip:
            idx = np.flip(idx)

        yield idx


def password_to_seed(password: str) -> int:
    """Converts password to seed for random generator.

    :param password: string password
    :type password: str
    :param seed: random seed for embedding simulator
    :type seed: int

    :Example:

    >>> seed = cl.tools.password_to_seed(cover_name)  # cover-specific seed
    >>> rng = np.random.default_rng(seed)
    """
    # compute SHA-256 digest of the password
    digest = hashlib.sha256(password.encode('utf-8')).hexdigest()
    # convert to seed
    seed = int(digest, base=16)
    return seed



"""

Author: Benedikt Lorch, Martin Benes
Affiliation: University of Innsbruck
"""

import numpy as np
import typing

from . import _costmap
from ._costmap import Change, Location
from ..simulate import _ternary


def simulate(
    cover: np.ndarray,
    alpha: float,
    *,
    modify: Change = Change.LSB_REPLACEMENT,
    locate: Location = None,
    permute: bool = None,  # deprecated
    cover_range: typing.Tuple[int] = (0, 255),
    n: int = None,
    e: float = 2,
    rhos: np.ndarray = None,
    wet_cost: float = 10**10,
    **kw,
) -> np.ndarray:
    """Simulates LSB steganography at an embedding rate into a cover,
    and returns stego.

    Allows both replacement and matching, sequential and permuted pass.
    LSB is well-known since early 1990's.

    The details of the methods are described in the
    `glossary <https://conseal.readthedocs.io/en/latest/glossary.html#least-significant-bit>`__.

    :param x0: cover image, in pixel or DCT domain,
        of arbitrary shape
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param alpha: embedding rate
        in bits per element
    :type alpha: float
    :param modify: modification strategy,
        replacement by default
    :type modify: :class:`Change`
    :param permute: permute the changes, otherwise sequential
    :type permute: bool
    :param cover_range: range of cover values,
        (0,255) by default
    :type cover_range: tuple
    :param n: cover size, used for DCT cover, number of elements by default
    :type n: int
    :param e: embedding efficiency
        in bits per change
    :type e: float
    :param wet_cost: wet cost for unembeddable coefficients
    :type wet_cost: float
    :param kw: remaining parameters passed to simulator
    :type kw: dict
    :return: stego image of the same meaning and shape as cover
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> x0 = np.array(Image.open('cover.pgm'))
    >>> x1 = cl.lsb.simulate(
    ...   cover=x0,
    ...   alpha=0.4,
    ...   modify=cl.LSB_MATCHING,
    ...   seed=12345)
    >>> Image.fromarray(x1).save('stego.pgm')

    The function can also take DCT coefficients

    >>> jpeg0 = jpeglib.read_dct('cover.jpeg')
    >>> jpeg1 = jpeg0.copy()
    >>> jpeg1.Y = cl.lsb.simulate(
    ...   cover=jpeg0.Y,
    ...   alpha=0.4,
    ...   modify=cl.LSB_MATCHING,
    ...   cover_range=(-1024, 1023),
    ...   seed=12345)
    >>> jpeg1.write_dct('stego.jpeg')

    Technically, the function can also simulate LSB for color images.
    This is the reason, why it is called `simulate` instead of `simulate_single_channel`.

    >>> x0 = np.array(Image.open('cover.ppm'))
    >>> x1 = cl.lsb.simulate(
    ...   cover=x0,
    ...   alpha=0.4,
    ...   modify=cl.LSB_MATCHING,
    ...   seed=12345)
    >>> Image.fromarray(x1).save('stego.ppm')
    """
    # compute probability
    ps, _ = _costmap.probability(
        cover=cover,
        alpha=alpha,
        modify=modify,
        locate=locate,
        permute=permute,
        cover_range=cover_range,
        n=n,
        e=e,
        rhos=rhos,
        wet_cost=wet_cost
    )
    # simulate
    delta = _ternary.simulate(
        ps=ps,
        **kw,
    )
    return cover + delta.astype(cover.dtype)



#!/usr/bin/env python3
"""
Implementation of S-UNIWARD steganographic embedding.

Code inspired by Holub's Matlab implementation: https://dde.binghamton.edu/download/stego_algorithms/
Code inspired by https://github.com/daniellerch/stegolab.

Universal Distortion Function for Steganography in an Arbitrary Domain.
V. Holub, J. Fridrich and T. Denemark. http://dde.binghamton.edu/vholub/pdf/EURASIP14_Universal_Distortion_Function_for_Steganography_in_an_Arbitrary_Domain.pdf
"""

import numpy as np
from scipy.signal import correlate2d
import typing

from .. import tools


def compute_cost(
    x0: np.ndarray,
    *,
    sigma: float = 1,
) -> np.ndarray:
    """Computes S-UNIWARD cost.

    :param x0: uncompressed (pixel) cover image,
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param sigma: stabilizing constant. Small sigmas make the embedding very sensitive to the image content. Large sigmas smooth out the embedding change probabilities.
    :type sigma: float
    :return: cost for +-1 change,
        of shape [height, width]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> # TODO
    """  # noqa: E501
    # Get 2D wavelet filters - Daubechies 8
    (high_pass_decomposition_filter, low_pass_decomposition_filter), filters = tools.spatial.daubechies8()
    num_filters = len(filters)

    pad_size = np.max([f.shape for f in filters])

    # Mirror-pad the spatial image. Extend image by the length of the filter in all dimensions.
    # "symmetric" means reflect, i.e. (d c b a | a b c d | d c b a)
    x0_padded = np.pad(x0, (pad_size, pad_size), 'symmetric')

    # Compute directional residual and suitability \xi for each filter
    xis = []
    for filter_idx in range(num_filters):
        # Compute residual

        # The original implementation uses a convolution but we take the correlation with the flipped kernel, which matches the Matlab alignment (otherwise, there is misalignment by 1 pixel).
        residual = correlate2d(x0_padded, np.flipud(np.fliplr(filters[filter_idx])), mode='same', boundary='fill', fillvalue=0)

        # Compute suitability
        xi = correlate2d(1. / (np.abs(residual) + sigma), np.abs(filters[filter_idx]), mode='same', boundary='fill', fillvalue=0)

        # Correct the suitability shift if filter size is even
        if filters[filter_idx].shape[0] % 2 == 0:
            xi = np.roll(xi, 1, axis=0)

        if filters[filter_idx].shape[1] % 2 == 0:
            xi = np.roll(xi, 1, axis=1)

        # Remove padding
        offset_y = (xi.shape[0] - x0.shape[0]) // 2
        offset_x = (xi.shape[1] - x0.shape[1]) // 2

        xi = xi[offset_y:-offset_y, offset_x:-offset_x]

        xis.append(xi)

    # Compute embedding costmap rho
    rho = xis[0] + xis[1] + xis[2]

    return rho


def compute_cost_adjusted(
    x0: np.ndarray,
    *,
    wet_cost: float = 10**8,
    **kw,
) -> typing.Tuple[np.ndarray, np.ndarray]:
    """Computes S-UNIWARD cost with wet-cost adjustments.

    :param x0: uncompressed (pixel) cover image,
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param wet_cost: wet cost for unembeddable coefficients
    :type wet_cost: float
    :return: costs for +1 and -1 changes,
        of shape [height, width]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> rhos = cl.suniward.compute_cost_adjusted(x0=x0)
    """
    assert len(x0.shape) == 2, 'single channel expected'

    # process input
    x0 = x0.astype('float64')

    # Calculate costmap
    rho = compute_cost(x0=x0, **kw)

    # Assign wet cost
    rho[np.isinf(rho) | np.isnan(rho) | (rho > wet_cost)] = wet_cost
    # rho[(rho > wet_cost) | np.isnan(rho)] = wet_cost

    # Do not embed +1 if the pixel has max value
    rho_p1 = np.copy(rho)
    rho_p1[x0 == 255] = wet_cost

    # Do not embed -1 if the pixel has min value
    rho_m1 = np.copy(rho)
    rho_m1[x0 == 0] = wet_cost

    # return costs
    return rho_p1, rho_m1


# def compute_cost_per_direction(
#     cover: np.ndarray,
#     sigma: float = 1,
# ) -> np.ndarray:
#     """
#     Produce UNIWARD distortion for given cover. Keep the three directions separate.

#     The first filter evaluates the smoothness in horizontal direction. The filter locates vertical edges.
#     The second filter evaluates the smoothness in vertical direction. The filter locates horizontal edges.
#     The third filter evaluates the smoothness in diagonal direction. The filter locates diagonal edges.

#     If the residual is high in direction d, the cost of embedding in direction d is low.

#     Args:
#         cover (np.ndarray): Cover image in spatial domain.
#         sigma (float): stabilizing constant. Small sigmas make the embedding very sensitive to the image content. Large sigmas smooth out the embedding change probabilities.
#     Returns:
#         costmap (np.ndarray): Same size as the cover image. The distortion for each of the three dimensions is stacked along the last axis.
#             First channel: Horizontal filter -> vertical edges
#             Second channel: Vertical filter -> horizontal edges
#             Third channel: Diagonal filter -> diagonal edges
#     """

#     # Get 2D wavelet filters - Daubechies 8
#     # 1D high-pass decomposition filter
#     # In the paper, the high-pass filter is denoted as g.
#     high_pass_decomposition_filter = np.array([-0.0544158422, 0.3128715909, -0.6756307363, 0.5853546837, 0.0158291053, -0.2840155430, -0.0004724846, 0.1287474266, 0.0173693010, -0.0440882539, -0.0139810279, 0.0087460940, 0.0048703530, -0.0003917404, -0.0006754494, -0.0001174768])
#     wavelet_filter_size = len(high_pass_decomposition_filter)

#     # 1D low-pass decomposition filter
#     # In the paper, the low-pass filter is denotes as h.
#     low_pass_decomposition_filter = np.power(-1, np.arange(len(high_pass_decomposition_filter))) * np.flip( high_pass_decomposition_filter)

#     # Stack filter kernels to shape [3, 16, 16]
#     # K^1 = h * g.T
#     # K^2 = g * h.T
#     # K^3 = g * g.T
#     filters = np.stack([
#         low_pass_decomposition_filter[:, None] * high_pass_decomposition_filter[None, :],
#         high_pass_decomposition_filter[:, None] * low_pass_decomposition_filter[None, :],
#         high_pass_decomposition_filter[:, None] * high_pass_decomposition_filter[None, :],
#     ], axis=0)
#     num_filters = len(filters)

#     pad_size = np.max([f.shape for f in filters])

#     # Mirror-pad the spatial image. Extend image by the length of the filter in all dimensions.
#     # "symmetric" means reflect, i.e. (d c b a | a b c d | d c b a)
#     cover_padded = np.pad(cover, (pad_size, pad_size), 'symmetric')

#     # Compute directional residual and suitability \xi for each filter
#     xis = []
#     for filter_idx in range(num_filters):
#         # Compute residual

#         # The original implementation uses a convolution but we take the correlation with the flipped kernel, which matches the Matlab alignment (otherwise, there is misalignment by 1 pixel).
#         residual = correlate2d(cover_padded, np.flipud(np.fliplr(filters[filter_idx])), mode='same', boundary='fill', fillvalue=0)

#         # Compute suitability
#         xi = correlate2d(1. / (np.abs(residual) + sigma), np.abs(filters[filter_idx]), mode='same', boundary='fill', fillvalue=0)

#         # Correct the suitability shift if filter size is even
#         if filters[filter_idx].shape[0] % 2 == 0:
#             xi = np.roll(xi, 1, axis=0)

#         if filters[filter_idx].shape[1] % 2 == 0:
#             xi = np.roll(xi, 1, axis=1)

#         # Remove padding
#         offset_y = (xi.shape[0] - cover.shape[0]) // 2
#         offset_x = (xi.shape[1] - cover.shape[1]) // 2

#         xi = xi[offset_y:-offset_y, offset_x:-offset_x]

#         xis.append(xi)

#     # Compute embedding costmap rho
#     rho_per_direction = np.stack(xis, axis=-1)

#     return rho_per_direction



#!/usr/bin/env python3
"""
Implementation of HILL steganographic embedding.

Code inspired by https://github.com/daniellerch/stegolab.

A New Cost Function for Spatial Image Steganography
B. Li, M. Wang, J. Huang, X. Li.
"""

import numpy as np
import scipy.signal
import typing

from .. import tools


def compute_cost(
    x0: np.ndarray,
) -> np.ndarray:
    """Computes HILL cost.

    :param x0: uncompressed (pixel) cover image
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :return: cost for +-1 change
        of shape [height, width]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> rho = cl.hill.compute_cost(x0=x0)
    """
    # high-pass filter
    H_KB = np.array([
        [-1, +2, -1],
        [+2, -4, +2],
        [-1, +2, -1]
    ], dtype='float32')
    I1 = scipy.signal.convolve2d(
        x0, H_KB,
        mode='same', boundary='symm',
    )

    # low-pass filter 1
    L1 = np.ones((3, 3), dtype='float32') / 3**2
    I2 = scipy.signal.convolve2d(
        np.abs(I1), L1,
        mode='same', boundary='symm',
    )

    # low-pass filter 2
    L2 = np.ones((15, 15), dtype='float32')/15**2
    I2[I2 < tools.EPS] = tools.EPS
    I3 = scipy.signal.convolve2d(
        1./(I2), L2,
        mode='same', boundary='symm',
    )

    #
    return I3


def compute_cost_adjusted(
    x0: np.ndarray,
    *,
    wet_cost: float = 1e10,
) -> typing.Tuple[np.ndarray]:
    """Computes HILL cost with wet-cost adjustments.

    :param x0: uncompressed (pixel) cover image
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param wet_cost: wet cost for unembeddable coefficients
    :type wet_cost: float
    :return: costs for +1 and -1 changes
        of shape [height, width]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> rhos = cl.hill.compute_cost_adjusted(x0=x0)
    """
    assert len(x0.shape) == 2, 'single channel expected'
    # process input
    x0 = x0.astype('float32')

    # Compute costmap
    rho = compute_cost(x0=x0)

    # Assign wet cost
    rho[np.isinf(rho) | np.isnan(rho) | (rho > wet_cost)] = wet_cost

    # Do not embed +1 if the pixel has max value
    rho_p1 = np.copy(rho)
    rho_p1[x0 >= 255] = wet_cost

    # Do not embed -1 if the pixel has min value
    rho_m1 = np.copy(rho)
    rho_m1[x0 <= 0] = wet_cost

    # return costs
    return rho_p1, rho_m1



"""Implementation of MiPOD steganographic embedding.

V. Sedighi, R. Cogranne and J. Fridrich
"Content-Adaptive Steganography by Minimizing Statistical Detectability"
IEEE Transactions on Information Forensics and Security
Vol. 11, no. 2, pp. 221-234, Feb. 2016.
http://dde.binghamton.edu/vsedighi/pdf/TIFS2015_Content_Adaptive_Steganography_by_Minimizing_Statistical_Detectability.pdf
Code inspired by Sideghi's Matlab implementation: https://dde.binghamton.edu/download/stego_algorithms/download/MiPOD_matlab.zip

Author: Benedikt Lorch, Martin Benes
Affiliation: University of Innsbruck
"""

import enum
import numpy as np
from typing import Tuple

from ._costmap import compute_cost
# from ..jmipod.simulate import ternary_probs

from . import _defs
from .. import tools


class Implementation(enum.Enum):
    """MiPOD implementation to choose from."""

    MiPOD_ORIGINAL = enum.auto()
    """Original MiPOD implementation by Remi Cogranne."""
    MiPOD_FIX_WET = enum.auto()
    """MiPOD implementation with wet-cost fixes."""


def probability(
    x0: np.ndarray,
    alpha: float,
) -> Tuple[Tuple[np.ndarray], float]:
    """Computes change probabilities for MiPOD embedding.

    :param x0: uncompressed (pixel) cover image
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param alpha: embedding rate
    :type alpha: float
    :return: tuple ((p_p1, p_m1), lmbda), where
        p_p1 is the probability of +1 change,
        p_m1 is the probability of -1 change, and
        lbda is the determined lambda.
    :rtype: tuple

    :Example:

    >>> (p_p1, p_m1), _ = cl.mipod.probability(
    ...     x0=x0,
    ...     alpha=.4)
    """
    fisher_information = compute_cost(x0=x0)

    # Absolute payload in nats
    payload = alpha * np.prod(fisher_information.shape) * np.log(2)

    # change rate per pixel/selection channel
    p_flat = _defs.ternary_probs(fisher_information, payload, max_num_iterations=60, excess_values_num_iter=20)

    p = p_flat.reshape(x0.shape, order="F")

    return (p, p), alpha * np.prod(fisher_information.shape)


def simulate_single_channel(
    x0: np.ndarray,
    alpha: float,
    *,
    implementation: Implementation = Implementation.MiPOD_FIX_WET,
    seed: int = None,
) -> np.ndarray:
    """Simulates MiPOD steganography.

    :param x0: uncompressed (pixel) cover image,
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param alpha: embedding rate,
        in bits per pixel
    :type alpha: float
    :param seed: random seed for embedding simulator
    :type seed: int
    :return: stego pixels,
        of shape [height, width, channels]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> x1 = cl.mipod.simulate_single_channel(
    ...     x0=x0,
    ...     alpha=.4,
    ...     seed=12345)
    """
    # rho is the fisher information
    rho = compute_cost(x0=x0)

    # Absolute payload in nats
    payload = alpha * np.prod(rho.shape) * np.log(2)

    # Beta is the change rate per pixel.
    # Beta is the selection channel.
    beta = _defs.ternary_probs(rho, payload, max_num_iterations=60, excess_values_num_iter=20)

    # The change rate for embedding +1 is beta and the change rate for embedding -1 is also beta.
    # Hence, the change of changing a pixel is 2 * beta.
    beta = 2 * beta

    rng = np.random.RandomState(seed)
    rand_change = rng.random_sample(x0.size)

    # Select cover elements to be modified by +- 1
    modify_pm_1 = rand_change < beta

    # Randomly select between +1 and -1 changes
    r = rng.random_sample(x0.size)
    delta_flat = tools.matlab_round(r[modify_pm_1]) * 2 - 1

    x1_flat = x0.copy().flatten(order="F").astype(int)
    x1_flat[modify_pm_1] += delta_flat.astype(int)
    x1 = x1_flat.reshape(x0.shape, order="F")

    # Take care of boundary cases
    if implementation == Implementation.MiPOD_ORIGINAL:
        x1[x1 > 255] = 253
        x1[x1 < 0] = 2
    elif implementation == Implementation.MiPOD_FIX_WET:
        x1[x1 > 255] = 254
        x1[x1 < 0] = 1
    else:
        raise NotImplementedError(f'unknown implementation {implementation}')

    return x1.astype(np.uint8)



"""

Author: Martin Benes
Affiliation: University of Innsbruck
"""

# costmap
from ._costmap import compute_cost_adjusted, Implementation

# simulate
from . import _simulate
from ._simulate import simulate_single_channel

# abbreviations of enum
EBS_ORIGINAL = Implementation.EBS_ORIGINAL
EBS_FIX_WET = Implementation.EBS_FIX_WET

__all__ = [
    '_costmap',
    '_simulate',
    'compute_cost_adjusted',
    'simulate_single_channel',
    'Implementation',
    'EBS_ORIGINAL',
    'EBS_FIX_WET',
]



"""
Implementation of the J-UNIWARD steganography method as described in

V. Holub, J. Fridrich, T. Denemark
"Universal distortion function for steganography in an arbitrary domain"
EURASIP Journal on Information Security, 2014
http://www.ws.binghamton.edu/fridrich/research/uniward-eurasip-final.pdf

Author: Benedikt Lorch, Martin Benes
Affiliation: University of Innsbruck

This implementation builds on the original Matlab implementation provided by the paper authors. Please find that license of the original implementation below.
-------------------------------------------------------------------------
Copyright (c) 2013 DDE Lab, Binghamton University, NY. All Rights Reserved.
Permission to use, copy, modify, and distribute this software for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that this copyright notice appears in all copies. The program is supplied "as is," without any accompanying services from DDE Lab. DDE Lab does not warrant the operation of the program will be uninterrupted or error-free. The end-user understands that the program was developed for research purposes and is advised not to rely exclusively on the program for any reason. In no event shall Binghamton University or DDE Lab be liable to any party for direct, indirect, special, incidental, or consequential damages, including lost profits, arising out of the use of this software. DDE Lab disclaims any warranties, and has no obligations to provide maintenance, support, updates, enhancements or modifications.
-------------------------------------------------------------------------
"""  # noqa: E501

import enum
import numpy as np
import scipy.signal
# from scipy.signal import correlate2d
import typing

from .. import tools


class Implementation(enum.Enum):
    """J-UNIWARD implementation to choose from."""

    JUNIWARD_ORIGINAL = enum.auto()
    """Original J-UNIWARD implementation."""
    JUNIWARD_FIX_OFF_BY_ONE = enum.auto()
    """J-UNIWARD implementation with fixed off-by-one error.

    See https://arxiv.org/pdf/2305.19776.pdf for more details.
    """


def compute_cost(
    x0: np.ndarray,
    qt: np.ndarray,
    *,
    sigma: float = 2**-6,
    dtype: np.dtype = np.float64,
    implementation: Implementation = Implementation.JUNIWARD_ORIGINAL,
) -> np.ndarray:
    """Compute the J-UNIWARD cost.

    :param x0: decompressed (pixel) image
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param qt: quantization table
        of shape [8, 8]
    :type qt: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param sigma: parameter controlling the content sensitivity, avoids division by zero but also controls the sensitivity to content.
    :type sigma: float
    :param dtype: data type to use for distortion computation, float64 by default
    :type dtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.dtype.html>`__
    :param implementation: choose J-UNIWARD implementation
    :type implementation: :class:`Implementation`
    :return: embedding cost
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> # TODO
    """  # noqa: E501
    assert len(x0.shape) == 2, "Expected grayscale image"
    height, width = x0.shape
    assert height % 8 == 0, "Expected height to be a multiple of 8"
    assert width % 8 == 0, "Expected width to be a multiple of 8"

    # Get 2D wavelet filters - Daubechies 8
    (high_pass_decomposition_filter, low_pass_decomposition_filter), filters = tools.spatial.daubechies8()
    wavelet_filter_size = len(high_pass_decomposition_filter)
    num_filters = len(filters)

    # Pre-compute impact in spatial domain when a JPEG coefficient is changed by 1
    spatial_impact = np.zeros((8, 8, 8, 8), dtype=dtype)
    for j in range(8):
        for i in range(8):
            # Simulate a single pixel change in the DCT domain
            y = np.zeros((8, 8), dtype=dtype)
            y[j, i] = 1
            spatial_impact[j, i] = tools.dct.idct2(y) * qt[j, i]

    # Pre-compute impact on wavelet coefficients when a JPEG coefficients is changed by 1
    # Changing one pixel will affect s x s wavelet coefficients, where s is the size of the 2D wavelet support.
    # Changing a DCT coefficient will affect a block of 8x8 pixels and therefore a block of (8 + s - 1) x (8 + s - 1) wavelet coefficients.
    change_window_len = 8 + wavelet_filter_size - 1
    assert change_window_len == 23, "Expected change window in Wavelet domain to have length 23"
    wavelet_impact = np.zeros((num_filters, 8, 8, change_window_len, change_window_len), dtype=dtype)
    for filter_idx in range(num_filters):
        for j in range(8):
            for i in range(8):
                # Output length of mode='full': N + M - 1
                wavelet_impact[filter_idx, j, i, :, :] = scipy.signal.correlate2d(
                    spatial_impact[j, i],
                    filters[filter_idx],
                    mode='full', boundary='fill', fillvalue=0)

    # Take absolute value, needed for later
    wavelet_impact = np.abs(wavelet_impact)

    # Create reference cover wavelet coefficients (LH, HL, HH)
    # Embedding should minimize their relative change.
    # Computation uses mirror-padding.
    pad_size = len(high_pass_decomposition_filter)

    # Mirror-pad the spatial image. Extend image by the length of the filter in all dimensions.
    # Example: Image [512, 512] -> [16 + 512 + 16, 16 + 512 + 16] = [544, 544]
    # "symmetric" means reflect, i.e. (d c b a | a b c d | d c b a)
    x0_padded = np.pad(x0, (pad_size, pad_size), mode='symmetric')

    # Compute the wavelet coefficients of the cover image
    reference_covers = []
    for filter_idx in range(num_filters):
        # Compute wavelet coefficients in the (filter_idx)-th subband
        # The resulting shape is [544, 544]
        rc = scipy.signal.correlate2d(
            x0_padded,
            filters[filter_idx],
            mode='same', boundary='fill', fillvalue=0,
        )

        # Crop as needed
        if implementation == Implementation.JUNIWARD_ORIGINAL:
            rc = rc[pad_size - 7:height + 8 + pad_size, pad_size - 7:width + 8 + pad_size]

        # My proposal
        elif implementation == Implementation.JUNIWARD_FIX_OFF_BY_ONE:
            # Can also be verified by setting the DCT coefficient to a very high value.
            rc = rc[pad_size - 8: height + 7 + pad_size, pad_size - 8:width + 7 + pad_size]

        else:
            raise NotImplementedError(f'unknonw implementation {implementation}')

        reference_covers.append(rc)

    #
    # Computation of costs
    #
    window_shape = (23, 23)
    stride = (8, 8)

    arr = reference_covers[0]
    # Compute the shape of the output array
    out_shape = ((arr.shape[0] - window_shape[0]) // stride[0] + 1,
                 (arr.shape[1] - window_shape[1]) // stride[1] + 1,
                 window_shape[0],
                 window_shape[1])

    # Compute the strides of the output array
    out_strides = (arr.strides[0] * stride[0],
                   arr.strides[1] * stride[1],
                   arr.strides[0],
                   arr.strides[1])

    # Use as_strided to create the sliding window view of the Wavelet coefficients
    arr_views = [np.lib.stride_tricks.as_strided(rc, shape=out_shape, strides=out_strides) for rc in reference_covers]

    # Prepare division by the Wavelet coefficients
    reciprocal_views = [np.reciprocal(np.abs(av) + sigma) for av in arr_views]

    # Divide over the [23, 23] subwindows without creating temporary arrays
    temp_xi_0 = np.einsum('ijkl,mnkl->mnij', wavelet_impact[0], reciprocal_views[0])
    temp_xi_1 = np.einsum('ijkl,mnkl->mnij', wavelet_impact[1], reciprocal_views[1])
    temp_xi_2 = np.einsum('ijkl,mnkl->mnij', wavelet_impact[2], reciprocal_views[2])

    # Sum over the three filters
    costs = temp_xi_0 + temp_xi_1 + temp_xi_2

    return costs


def compute_cost_adjusted(
    x0: np.ndarray,
    y0: np.ndarray,
    qt: np.ndarray,
    *,
    dtype: np.dtype = np.float64,
    implementation: Implementation = Implementation.JUNIWARD_ORIGINAL,
    wet_cost: float = 10**13,
    avoid_saturated: bool = False,
) -> typing.Tuple[np.ndarray, np.ndarray]:
    """Compute the adjusted J-UNIWARD cost for a ternary embedding.

    :param x0: decompressed (pixel) image,
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param y0: quantized cover DCT coefficients,
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :type y0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param qt: quantization table,
        of shape [8, 8]
    :type qt: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param dtype: data type to use for distortion computation, float64 by default
    :type dtype: np.dtype
    :param implementation: choose J-UNIWARD implementation
    :type implementation: :class:`Implementation`
    :param wet_cost: cost for unembeddable coefficients
    :type wet_cost: float
    :param avoid_saturated: hard-sets blocks with saturated pixels to wet
    :type avoid_saturated: bool
    :return: embedding costs of +1 and -1 changes,
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :rtype: tuple of `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> rho_p1, rho_m1 = cl.juniward._costmap.compute_cost_adjusted(
    ...     y0=jpeg0.Y,      # DCT
    ...     qt=jpeg0.qt[0],  # QT
    ...     x0=im0.spatial)  # decompressed spatial
    >>> im_dct.Y += cl.simulate.ternary(
    ...     rhos=(rho_p1, rho_m1),  # distortion of +1 and -1 change
    ...     alpha=0.4,              # alpha
    ...     n=im_dct.Y.size,        # cover size
    ...     seed=12345)             # seed
    """  # noqa: E501
    # Count number of embeddable DCT coefficients
    assert tools.dct.nzAC(y0) > 0, 'Expected non-zero AC coefficients'
    # num_non_zero_AC_coeffs = tools.dct.nzAC(y0)
    # if num_non_zero_AC_coeffs == 0:
    #     raise ValueError('Expected non-zero AC coefficients')

    # Compute costmap
    rho = compute_cost(
        x0=x0,
        qt=qt,
        dtype=dtype,
        implementation=implementation,
    )

    # Assign wet cost
    rho[np.isinf(rho) | np.isnan(rho) | (rho > wet_cost)] = wet_cost
    if avoid_saturated:
        rho[(tools.jpegio_to_jpeglib(x0) == 0).any(axis=(2, 3))] = wet_cost
        rho[(tools.jpegio_to_jpeglib(x0) == 255).any(axis=(2, 3))] = wet_cost

    # Do not embed +1 if the DCT coefficient has max value
    rho_p1 = np.copy(rho)
    rho_p1[y0 >= 1023] = wet_cost

    # Do not embed -1 if the DCT coefficient has min value
    rho_m1 = np.copy(rho)
    rho_m1[y0 <= -1023] = wet_cost

    return rho_p1, rho_m1


def evaluate_cost(
    x0: np.ndarray,
    x1: np.ndarray,
    sigma: float = 2**-6,
) -> float:
    """Compute the J-UNIWARD cost between x0 and x1.

    The UNIWARD distortion function is the sum of relative changes of all wavelet coefficients with respect to the cover image.

    Note that both images are padded symmetrically with the cover image margins.
    This is necessary to match the Matlab implementation.

    :param x0: cover image of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param x1: stego image of shape [height, width]
    :type x1: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :return: distortion as scalar value
    :rtype: float
    """
    # # Get 2D wavelet filters - Daubechies 8
    # # 1D high-pass decomposition filter
    # # In the paper, the high-pass filter is denoted as g.
    # high_pass_decomposition_filter = np.array([
    #     -0.0544158422, 0.3128715909, -0.6756307363, 0.5853546837,
    #     0.0158291053, -0.2840155430, -0.0004724846, 0.1287474266,
    #     0.0173693010, -0.0440882539, -0.0139810279, 0.0087460940,
    #     0.0048703530, -0.0003917404, -0.0006754494, -0.0001174768,
    # ])

    # # 1D low-pass decomposition filter
    # # In the paper, the low-pass filter is denotes as h.
    # low_pass_decomposition_filter = np.power(-1, np.arange(len(high_pass_decomposition_filter))) * np.flip(high_pass_decomposition_filter)

    # # Stack filter kernels to shape [3, 16, 16]
    # # K^1 = h * g.T
    # # K^2 = g * h.T
    # # K^3 = g * g.T
    # filters = np.stack([
    #     low_pass_decomposition_filter[:, None] * high_pass_decomposition_filter[None, :],
    #     high_pass_decomposition_filter[:, None] * low_pass_decomposition_filter[None, :],
    #     high_pass_decomposition_filter[:, None] * high_pass_decomposition_filter[None, :],
    # ], axis=0)

    (high_pass_decomposition_filter, low_pass_decomposition_filter), filters = tools.spatial.daubechies8()
    num_filters = len(filters)

    # Pad X and Y by 16 pixels in all directions
    pad_size = len(high_pass_decomposition_filter)
    x0_padded = np.pad(x0, (pad_size, pad_size), mode='symmetric')

    # Pad Y with the same pixels as X.
    # Otherwise, changes between Y and X would be also be present the padded region.
    x1_padded = np.pad(x0, (pad_size, pad_size), mode='symmetric')
    x1_padded[pad_size:-pad_size, pad_size:-pad_size] = x1

    w0 = []
    w1 = []
    for filter_idx in range(num_filters):
        # Compute wavelet coefficients in the (filter_idx)-th subband
        w0_k = scipy.signal.correlate2d(
            x0_padded,
            filters[filter_idx],
            mode='same', boundary='fill', fillvalue=0)
        w1_k = scipy.signal.correlate2d(
            x1_padded,
            filters[filter_idx],
            mode='same', boundary='fill', fillvalue=0)

        # For completeness, this is identical to
        # assert np.allclose(
        #     W_X_k[pad_size:-pad_size, pad_size:-pad_size],
        #     correlate2d(X, filters[filter_idx], mode='same', boundary='symmetric')
        # )
        # The same check will fail for Y because we padded Y_padded with the boundaries of X.

        # To comply with Eq. 3, we would have to crop the padding.
        # However, the padding needs to be retained to match the Matlab implementation.
        w0.append(w0_k)
        w1.append(w1_k)

    # Stack filtered images.
    # The resulting shape is [num_filters, 16 + height + 16 , 16 + width + 16].
    w0 = np.stack(w0, axis=0)
    w1 = np.stack(w1, axis=0)

    # Fraction in Eq. 3
    per_pixel_distortion = np.abs(w0 - w1) / (np.abs(w0) + sigma)

    # Eq. 3: Sum over filters, height, and width
    distortion = np.sum(per_pixel_distortion)

    return distortion



"""

Author: Martin Benes
Affiliation: University of Innsbruck
"""

import enum
import numpy as np
import scipy.signal
from typing import Tuple

from .. import mipod
from .common import EPS


class Attacker(enum.Enum):
    """Type of atacker."""

    ATTACKER_OMNISCIENT = enum.auto()
    """Omniscient attacker with SCA."""
    ATTACKER_INDIFFERENT = enum.auto()
    """Indifferent attacker without SCA."""


def estimate_variance(
    x: np.ndarray,
) -> np.ndarray:
    """
    Estimates the pixels' variance in 3x3 pixel neighborhood as

    V[X] = E[X^2] - (E[X])^2

    :param x: image
    :type x: np.ndarray
    :param block_size: size of the local neighborhood, denoted as p in the paper. See Fig. 2.
        Small p: extreme content adaptivity;
        Medium p: Medium content adaptivity;
        Large p: Low content adaptivity.
    :type block_size: int
    :param degree: degree of the polynomial
    :type degree: int
    :return: estimated variance per pixel
    :rtype:
    """
    """"""
    x = x.astype('float32')

    # local mean
    kernel = np.ones((3, 3), dtype='float')
    R = scipy.signal.convolve2d(np.ones_like(x), kernel, mode='same', boundary='fill')
    Ex = scipy.signal.convolve2d(x, kernel, mode='same', boundary='fill') / R

    # local variance
    Ex2 = scipy.signal.convolve2d(x**2, kernel, mode='same', boundary='fill') / R
    Vx = Ex2 - Ex**2
    return Vx


def attack(
    x0: np.ndarray,
    ps: Tuple[np.ndarray],
    *,
    clip: float = EPS,
    attacker: Attacker = Attacker.ATTACKER_OMNISCIENT,
) -> float:
    """Likelihood ratio test with cover x0 and assumed change rates.

    The method assumes a Gaussian cover model.
    In practice, it is used to quickly benchmark steganographic methods,
    without need to train an actual classifier.

    It was introduced in
    Sedighi, Cogranne, Fridrich.
    Content-Adaptive Steganography by Minimizing Statistical Detectability.
    IEEE TIFS, 2016.

    :param x0: uncompressed (pixel) cover image
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param ps: probability tensors for changes
        of an arbitrary shape
    :type ps: tuple of `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param clip: bottom clip of variance to ensure numerical stability,
        EPS by default
    :type clip: float
    :param attacker: assumed attacker,
        omniscient attacker (knowing SCA) by default
    :type attacker: :class:`Attacker`
    :return: value of the deflection coefficient
    :rtype: float

    :Example:

    >>> rho_p1, rho_m1 = cl.hill.compute_cost_adjusted(x0)
    >>> (p_p1, p_m1), lbda = cl.simulate._ternary.probability(
    ...     rhos=(rho_p1, rho_m1),
    ...     alpha=.1,
    ...     n=x0.size)
    >>> defl_hill = cl.tools.lrt.attack(x0, (p_p1, p_m1), clip=1e-3)
    """
    assert len(x0.shape) == 2, 'grayscale expected'
    x0 = x0.astype('float32')
    # ternary to binary
    p = np.sum(ps, axis=0)

    # estimate variance (with MiPOD estimator)
    # Vx0 = estimate_variance(x0)  # worse
    v0 = x0 - mipod.wiener2(x0, kernel_size=(2, 2))
    Vx0 = mipod.estimate_variance(v0, block_size=3, degree=3)

    # fisher information
    Vx0 = np.clip(Vx0, clip, None)  # clip FI in flat areas
    fi = 1. / Vx0**2
    fi[np.isinf(fi)] = 0  # ignores infinities

    # deflection coefficient
    # Eq. 11
    if attacker == Attacker.ATTACKER_OMNISCIENT:
        lr = np.sqrt(2) * np.sqrt(np.nansum(fi * p**2))
    # Eq. 12
    elif attacker == Attacker.ATTACKER_INDIFFERENT:
        lr = np.sqrt(2) * np.nansum(fi * p) / np.sqrt(np.nansum(fi))
    else:
        raise NotImplementedError(f'unknown attacker {attacker}')
    return float(lr)



"""

Author: Martin Benes
Affiliation: University of Innsbruck
"""

import numpy as np
from scipy.signal import convolve2d


def compute_cost(x0: np.ndarray) -> np.ndarray:
    """Computes a adversarial cost against WS steganalysis.

    It transforms the WS residual using the exponential function.

    Intended for selection-channel steganography for simple comparison
    against vanilla LSBR performance.

    :param x0: uncompressed (pixel) cover image
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :returns: cost against WS
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    """
    # flip
    x0_bar = x0 ^ 1
    x0 = x0.astype('float32')

    # predict
    f_kb = np.array([
        [-1, +2, -1],
        [+2,  0, +2],
        [-1, +2, -1],
    ], dtype='float32') / 4.
    x0_hat = convolve2d(
        x0, f_kb,
        mode='same', boundary='symm'
    )

    # calculate local WS residuals
    betas_hat = (x0 - x0_bar) * (x0 - x0_hat)

    # convert to distortion
    rhos = np.exp(-np.abs(betas_hat))
    """This is likely not optimal, but I had little time to test something better."""
    return rhos


def compute_cost_adjusted(
    x0,
    *,
    wet_cost: float = 1e10,
) -> np.ndarray:
    """Computes a adversarial cost against WS steganalysis.

    It transforms the WS residual using the exponential function.

    :param x0: uncompressed (pixel) cover image
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param wet_cost: wet cost for unembeddable coefficients
    :type wet_cost: float
    :returns: cost against WS
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    """
    # compute cost
    rho = compute_cost(x0)


    # Assign wet cost
    rho[np.isinf(rho) | np.isnan(rho) | (rho > wet_cost)] = wet_cost

    # Do not embed +1 if the pixel has max value
    rho_p1 = np.copy(rho)
    rho_p1[x0 >= 255] = wet_cost

    # Do not embed -1 if the pixel has min value
    rho_m1 = np.copy(rho)
    rho_m1[x0 <= 0] = wet_cost

    # return costs
    return rho_p1, rho_m1



"""

Author: Martin Benes, Benedikt Lorch
Affiliation: University of Innsbruck
"""

# globals
from .common import entropy, inv_entropy, _entropy, EPS, matlab_round
from .dct import nzAC, AC, jpegio_to_jpeglib, jpeglib_to_jpegio
from .decompress import decompress_channel
from .permute import blocks, iterate_ac, iterate, password_to_seed

# modules
from . import dct
from . import spatial
from . import lrt

ATTACKER_INDIFFERENT = lrt.Attacker.ATTACKER_INDIFFERENT
ATTACKER_OMNISCIENT = lrt.Attacker.ATTACKER_OMNISCIENT

__all__ = [
    'entropy',
    'inv_entropy',
    '_entropy',
    'EPS',
    'nzAC',
    'AC',
    'dct',
    'spatial',
]



"""
Implementation of simulated LSB steganography.

Unlike other methods, here we do not have a reference code,
and test visually and using known attacks and weaknesses of LSB.

Author: Martin Benes
Affiliation: University of Innsbruck
"""  # noqa: E501

import enum
import numpy as np
import typing
import warnings

from . import _selection_channel
from .. import simulate
from .. import tools


class Change(enum.Enum):
    """Modify strategy for LSB steganography."""

    LSB_REPLACEMENT = enum.auto()
    """LSB replacement."""
    LSB_MATCHING = enum.auto()
    """LSB matching."""


class Location(enum.Enum):
    """Location of embedding changes."""

    LOCATION_PERMUTED = enum.auto()
    """Permuted embedding."""
    LOCATION_SEQUENTIAL = enum.auto()
    """Sequential embedding."""
    LOCATION_SELECTED = enum.auto()
    """Selection channel."""


def compute_cost(
    cover: np.ndarray,
    *,
    modify: Change = Change.LSB_REPLACEMENT,
    wet_cost: float = 10**10,
) -> np.ndarray:
    """Returns LSB cost.

    Provides unified interface with cost-based embeddings.

    :param cover: cover image, in pixel or DCT domain, of arbitrary shape
    :type cover: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param modify: modification strategy, replacement by default
    :type modify: :class:`Change`
    :param cover_range: Range of cover values, (0,255) by default.
    :type cover_range: tuple
    :param wet_cost: wet cost for unembeddable elements
    :type wet_cost: float
    :return: cost of the same shape as cover
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> # TODO
    """
    rho_p1 = np.ones(cover.shape)
    rho_m1 = np.ones(cover.shape)

    # LSB replacement
    if modify == Change.LSB_REPLACEMENT:
        rho_p1[cover % 2 != 0] = wet_cost
        rho_m1[cover % 2 == 0] = wet_cost

    # LSB matching
    elif modify == Change.LSB_MATCHING:
        pass

    else:
        raise NotImplementedError(
            f'unknown modify strategy {modify}'
        )

    return rho_p1, rho_m1


def compute_cost_adjusted(
    cover: np.ndarray,
    *,
    modify: Change = Change.LSB_REPLACEMENT,
    cover_range: typing.Tuple[int] = (0, 255),
    wet_cost: float = 10**10,
) -> np.ndarray:
    """Returns LSB distortion.

    Provides unified interface with distortion-based embeddings.

    :param cover: cover image, in pixel or DCT domain, of arbitrary shape
    :type cover: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param modify: modification strategy, replacement by default
    :type modify: :class:`Change`
    :param cover_range: Range of cover values, (0,255) by default.
    :type cover_range: tuple
    :param wet_cost: wet cost for unembeddable elements
    :type wet_cost: float
    :return: distortion of the same shape as cover
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> rhos = cl.lsb.compute_cost_adjusted(x0)
    >>> x1 = x0 + cl.simulate.ternary(rhos=rhos, alpha=.4, seed=12345)
    """
    # compute cost
    rho_p1, rho_m1 = compute_cost(
        cover=cover,
        modify=modify,
        wet_cost=wet_cost,
    )

    # sanitize range for LSB matching
    if modify == Change.LSB_MATCHING:
        cover_min, cover_max = cover_range
        where_min = cover == cover_min
        where_max = cover == cover_max
        rho_p1[where_min], rho_m1[where_min] = 0, rho_p1[where_min]
        rho_m1[where_max], rho_p1[where_max] = 0, rho_m1[where_max]

    return rho_p1, rho_m1


def probability(
    cover: np.ndarray,
    alpha: float,
    *,
    modify: Change = Change.LSB_REPLACEMENT,
    locate: Location = None,
    permute: bool = None,  # deprecated
    cover_range: typing.Tuple[int] = (0, 255),
    n: int = None,
    rhos: np.array = None,
    e: float = 2,
    wet_cost: float = 10**10,
) -> np.ndarray:
    """Returns LSB probability map for consequent simulation.

    :param cover: cover image
        of arbitrary shape
        in pixel or DCT domain
    :type cover: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param alpha: embedding rate
        in bits per pixel
    :type alpha: float
    :param modify: modification strategy, replacement by default
    :type modify: :class:`Change`
    :param permute: Permute the changes, otherwise sequential
    :type permute: bool
    :param cover_range: Range of cover values, (0,255) by default.
    :type cover_range: tuple
    :param n: cover size, used for DCT cover, number of elements by default
    :type n: int
    :param e: embedding efficiency
        in bits per change
    :type e: float
    :param rhos: costs for the selection channel
    :type rhos: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param wet_cost: wet cost for unembeddable elements
    :type wet_cost: float
    :return: probability map of the same shape as cover
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> ps, _ = cl.lsb.probability(x0, alpha=.4)
    >>> x1 = x0 + cl.simulate._ternary.simulate(ps=ps, seed=12345)
    """
    if permute is not None:
        warnings.warn(
            'permuted is deprecated, use locate instead',
            DeprecationWarning,
        )
        if permute:
            locate = Location.LOCATION_PERMUTED
        else:
            locate = Location.LOCATION_SEQUENTIAL
    elif locate is None:
        locate = Location.LOCATION_PERMUTED

    #
    if n is None:
        n = cover.size
    if e is None:
        e = alpha / tools.inv_entropy(alpha)

    p_p1 = np.zeros(cover.shape, dtype='double')
    p_m1 = np.zeros(cover.shape, dtype='double')
    if alpha > 0:
        # number of changes
        involved_elements = int(np.ceil(alpha * n))
        changes = int(np.ceil(involved_elements / e))
        change_rate = changes / n

        # permutative straddling
        if locate == Location.LOCATION_PERMUTED:
            p = np.ones(cover.shape, dtype=float) * change_rate
        # sequential embedding
        elif locate == Location.LOCATION_SEQUENTIAL:
            p = np.reshape(
                [1/e]*involved_elements + [0]*(n - involved_elements),
                cover.shape
            )
        # selection channel
        elif locate == Location.LOCATION_SELECTED:
            assert rhos is not None, 'rhos were not given'
            p = _selection_channel.probability(rhos=rhos, alpha=alpha, e=e)

        # LSB replacement
        if modify == Change.LSB_REPLACEMENT:
            p_p1[cover % 2 == 0] = p[cover % 2 == 0]
            p_m1[cover % 2 != 0] = p[cover % 2 != 0]

        # LSB matching
        elif modify == Change.LSB_MATCHING:
            p_p1[:] = p/2
            p_m1[:] = p/2
            # sanitize min and max
            cover_min, cover_max = cover_range
            where_min = cover == cover_min
            where_max = cover == cover_max
            p_m1[where_min], p_p1[where_min] = 0, p[where_min]
            p_p1[where_max], p_m1[where_max] = 0, p[where_max]

        else:
            raise NotImplementedError(
                f'no modify strategy {modify}'
            )

    return (p_p1, p_m1), None


def average_payload(*args, **kw):
    """Objective function for using LSB together with simulator.

    It sets a constant efficiency over all the embedding rates to 2.
    Other methods embed at the bound.
    """
    return simulate._ternary.average_payload(*args, e=2, **kw)



"""
Implementation of the nsF5 steganography method as described in

J. Fridrich, T. Pevny, and J. Kodovsky.
"Statistically undetectable JPEG steganography: Dead ends, challenges, and opportunities"
Multimedia & Security, 2007
http://dde.binghamton.edu/kodovsky/pdf/Fri07-ACM.pdf

Author: Benedikt Lorch, Martin Benes
Affiliation: University of Innsbruck

This implementation is derived from the original Matlab implementation provided by the paper authors: https://dde.binghamton.edu/download/stego_algorithms/
"""

import numpy as np

from .. import tools


def average_payload(
    y0: np.ndarray,
    y1: np.ndarray,
) -> float:
    """Estimates payload [bpnzac] embedded in the stego with nsF5.

    :param y0: quantized cover DCT coefficients
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :type y0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param y1: quantized stego DCT coefficients
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :type y1: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :return: estimated embedding rate
    :rtype: float

    :Example:

    >>> # TODO
    """
    num_changes = (y0 != y1).sum()
    nzAC = tools.dct.nzAC(y0)
    return tools.H(num_changes / nzAC)


def simulate_single_channel(
    y0: np.ndarray,
    alpha: float,
    *,
    add_shrinkage: bool = False,
    e: float = None,
    seed: int = None,
) -> np.ndarray:
    """Simulates nsF5 embedding at an embedding rate into single-channel cover and returns stego.

    nsF5 was introduced in
    J. Fridrich, et al. Statistically undetectable JPEG steganography: Dead ends, challenges, and opportunities.
    ACM MMSec, 2007.

    The details of the methods are described in the
    `glossary <https://conseal.readthedocs.io/en/latest/glossary.html#non-shrinkage-f5-nsf5>`__.

    :param y0: quantized cover DCT coefficients
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :type y0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param alpha: embedding rate
    :type alpha: float
    :param e: embedding efficiency
        in bits per change
    :param seed: random seed for embedding simulator
    :type seed: int
    :return: quantized stego DCT coefficients,
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> im_dct.Y = cl.nsF5.simulate_single_channel(
    ...   y0=im_dct.Y,  # DCT
    ...   alpha=0.4,  # alpha
    ...   seed=12345)  # seed
    """
    assert len(y0.shape) == 4, "Expected DCT coefficients to have 4 dimensions"
    assert y0.shape[2] == y0.shape[3] == 8, "Expected blocks of size 8x8"

    # No embedding
    if np.isclose(alpha, 0):
        return y0

    # Compute bound on embedding efficiency
    if e is None:
        e = alpha / tools.inv_entropy(alpha)

    # Number of nonzero DCT AC coefficients
    nzAC = tools.dct.nzAC(y0)
    assert nzAC > 0, 'there are no coefficients to embed to'

    # Calculate change rate on the bound
    beta = alpha / e

    # Introduce shrinkage
    if add_shrinkage:
        p_pm1 = np.mean(np.abs(y0) == 1)  # probability of 1 or -1
        beta = beta / (1 - p_pm1)  # adjust change rate
        """
        beta' = beta + beta * p + beta * p**2 + ...
        beta' = beta + beta * (sum_{i=1:infty} p**i)
        beta' = beta + beta * p / (1 - p)
        beta' = beta / (1 - p)
        """

    # Number of changes
    num_changes = int(np.ceil(beta * nzAC))

    # # Number of changes nsF5 would make on bound
    # num_changes = int(np.ceil(alpha * nzAC / e))

    # Rearrange DCT coefficients to image shape in order to match the Matlab implementation
    num_vertical_blocks, num_horizontal_blocks, _, _ = y0.shape

    # Mask of all nonzero DCT coefficients in the image
    changeable = y0 != 0

    # Do not embed into DC modes
    changeable[:, :, 0, 0] = False

    # Convert to 2D
    changeable_2d = changeable.transpose((0, 2, 1, 3)).reshape((num_vertical_blocks * 8, num_horizontal_blocks * 8))

    # Indexes of changeable coefficients
    indices_2d = np.stack(np.where(changeable_2d), axis=1)

    # Permutative straddling
    # Initialize PRNG using given seed
    rng = np.random.RandomState(seed)

    # Create a pseudorandom walk over nonzero AC coefficients
    permutation = rng.permutation(nzAC)

    # Permute indices
    permuted_indices_2d = indices_2d[permutation]

    # # Temporarily save the permutation
    # store_permutation = False
    # if store_permutation:
    #     from scipy.io import savemat

    #     # Create an array to hold the order of coefficients used by Matlab (only count the changeable coefficients)
    #     indices_matlab = np.zeros((num_vertical_blocks * 8, num_horizontal_blocks * 8), dtype=int)

    #     # Because Matlab uses column-major order, reorder our coordinates by the y-dimension
    #     indices_2d_reordered = indices_2d[np.lexsort((indices_2d[:, 0], indices_2d[:, 1]))]

    #     # Fill index array sequentially
    #     indices_matlab[indices_2d_reordered[:, 0], indices_2d_reordered[:, 1]] = np.arange(nzAC)

    #     # Retrieve the permutation that Matlab will apply
    #     permutation_matlab = indices_matlab[permuted_indices_2d[:, 0], permuted_indices_2d[:, 1]]

    #     # Store the permutation to file
    #     savemat(f"/tmp/random_permutation_seed_{seed}_nzAC_{nzAC}.mat", {"permutation": permutation_matlab})

    # Flatten cover DCT coefficients
    y0_2d = y0.transpose((0, 2, 1, 3)).reshape((num_vertical_blocks * 8, num_horizontal_blocks * 8))

    # # Re-embed zeros (shrinkage)
    # if add_shrinkage:
    #     start = 0
    #     num_zeros = (
    #         np.abs(y0_2d[tuple(permuted_indices_2d[start:num_changes].T)]) == 1
    #     ).sum()
    #     while num_zeros > 0:
    #         # Move segment
    #         start = num_changes
    #         num_changes += num_zeros
    #         # Number of zeros in new segment
    #         num_zeros = (
    #             np.abs(y0_2d[tuple(permuted_indices_2d[start:num_changes].T)]) == 1
    #         ).sum()

    # Coefficients to be changed
    to_be_changed_2d = permuted_indices_2d[:num_changes]

    # Buffer containing the changes to the cover image
    delta_2d = np.zeros(y0_2d.shape, dtype=np.int8)

    # Decrease the absolute value of the coefficients to be changed
    delta_2d[to_be_changed_2d[:, 0], to_be_changed_2d[:, 1]] = -np.sign(y0_2d[to_be_changed_2d[:, 0], to_be_changed_2d[:, 1]])

    # Reshape delta to the original shape
    delta = delta_2d.reshape((num_vertical_blocks, 8, num_horizontal_blocks, 8)).transpose((0, 2, 1, 3))
    # return delta
    return y0 + delta


__all__ = ['simulate_single_channel']



"""

Author: Benedikt Lorch, Martin Benes
Affiliation: University of Innsbruck
"""

import numpy as np

from ._costmap import compute_cost_adjusted, Implementation
from .. import simulate
from .. import tools


def simulate_single_channel(
    x0: np.ndarray,
    y0: np.ndarray,
    qt: np.ndarray,
    alpha: float,
    *,
    wet_cost: float = 10**13,
    dtype: np.dtype = np.float64,
    implementation: Implementation = Implementation.JUNIWARD_ORIGINAL,
    generator: str = None,
    seed: int = None,
) -> np.ndarray:
    """Simulates J-UNIWARD embedding at an embedding rate into single-channel cover,
    and returns stego.

    J-UNIWARD was introduced in
    V. Holub, et al. Universal distortion function for steganography in an arbitrary domain.
    EURASIP JIS, 2014.

    The details of the methods are described in the
    `glossary <https://conseal.readthedocs.io/en/latest/glossary.html#jpeg-universal-wavelet-relative-distortion-j-uniward>`__.

    :param x0: decompressed (pixel) cover image
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param y0: quantized cover DCT coefficients
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :type y0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param qt: quantization table
        of shape [8, 8]
    :type qt: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param alpha: embedding rate
        in bits per nzAC coefficient
    :type alpha: float
    :param wet_cost: wet cost for unembeddable coefficients
    :type wet_cost: float
    :param dtype: data type to use for distortion computation,
        float64 by default
    :type dtype: `np.dtype <https://numpy.org/doc/stable/reference/generated/numpy.dtype.html>`__
    :param implementation: choose J-UNIWARD implementation
    :type implementation: :class:`Implementation`
    :param generator: type of PRNG used by embedding simulator
    :type generator: `numpy.random.Generator <https://numpy.org/doc/stable/reference/random/generator.html>`__
    :param seed: random seed for embedding simulator
    :type seed: int
    :return: quantized stego DCT coefficients
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> jpeg1.Y = cl.juniward.simulate_single_channel(
    ...   y0=jpeg0.Y,  # DCT
    ...   qt=jpeg0.qt[0],  # QT
    ...   x0=im0.spatial[..., 0],  # decompressed
    ...   alpha=0.4,  # embedding rate
    ...   seed=12345)  # seed
    """
    # Count number of embeddable DCT coefficients
    nzAC = tools.dct.nzAC(y0)

    if nzAC == 0:
        raise ValueError('Expected non-zero AC coefficients')

    # Compute cost for embedding into the quantized DCT coefficients
    # of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    rho_p1, rho_m1 = compute_cost_adjusted(
        x0=x0,
        y0=y0,
        qt=qt,
        dtype=dtype,
        implementation=implementation,
        wet_cost=wet_cost,
    )

    # Rearrange from 4D to 2D
    rho_p1_2d = tools.dct.jpeglib_to_jpegio(rho_p1)
    rho_m1_2d = tools.dct.jpeglib_to_jpegio(rho_m1)

    # STC simulation
    delta_2d = simulate.ternary(
        rhos=(rho_p1_2d, rho_m1_2d),
        alpha=alpha,
        n=nzAC,
        generator=generator,
        seed=seed,
    )

    # Convert from 2D to 4D
    delta = tools.dct.jpegio_to_jpeglib(delta_2d)

    # stego = cover + stego noise
    return y0 + delta



#!/usr/bin/env python3
"""
Implementation of HUGO steganographic embedding.

Code inspired by Holub's Matlab implementation: https://dde.binghamton.edu/download/stego_algorithms/

Using High-Dimensional Image Models to Perform Highly Undetectable Steganography.
T. Pevny, T. Filler and P. Bas. https://hal.science/hal-00541353/document
"""


from numba import jit, prange
import numpy as np
import typing


@jit(nopython=True, fastmath=True, nogil=False, cache=True, parallel=False)
def GetLocalPotential(
    c_res: np.ndarray,
    s_res: np.ndarray,
    sigma: float,
    gamma: float,
) -> float:
    """"""
    c_w = (sigma + np.sqrt(np.sum(c_res**2)))**-gamma
    s_w = (sigma + np.sqrt(np.sum(s_res**2)))**-gamma
    Vc = (c_w + s_w)
    return Vc


@jit(nopython=True, fastmath=True, nogil=False, cache=True, parallel=False)
def GetLocalDistortion(
    c_res: np.ndarray,
    s_res: np.ndarray,
    sigma: float,
    gamma: float,
) -> float:
    """"""
    D = 0.
    assert len(c_res) == 6, 'invalid c_res size'
    assert len(s_res) == 6, 'invalid S_resVect size'
    D += GetLocalPotential(c_res[0:3], s_res[0:3], sigma, gamma)
    D += GetLocalPotential(c_res[1:4], s_res[1:4], sigma, gamma)
    D += GetLocalPotential(c_res[2:5], s_res[2:5], sigma, gamma)
    D += GetLocalPotential(c_res[3:6], s_res[3:6], sigma, gamma)
    return D


@jit(nopython=True, fastmath=True, nogil=False, cache=True, parallel=False)
def compute_cost(
    x0: np.ndarray,
    x0_padded: np.ndarray,
    sigma: float = 1,
    gamma: float = 1,
) -> typing.Tuple[np.ndarray]:
    """Computes HUGO cost.

    Unlike most other distortions, HUGO is truly directional.

    :param x0: uncompressed (pixel) cover image
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param x0_padded:
    :type x0_padded: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param sigma: parameter
    :type sigma: float
    :param gamma: parameter
    :type gamma: float
    :return: cost of the same shape as cover
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> # TODO
    """
    # embedding costs initialization
    response_p1 = np.array([0, 0, -1, +1, 0, 0])

    # declare costs
    rho_p1 = np.zeros(x0.shape, dtype='double')
    rho_m1 = np.zeros(x0.shape, dtype='double')

    # create residual
    d0_h = x0_padded[:, :-1] - x0_padded[:, 1:]
    d0_v = x0_padded[:-1, :] - x0_padded[1:, :]
    d0_d = x0_padded[:-1, :-1] - x0_padded[1:, 1:]
    d0_m = x0_padded[:-1, 1:] - x0_padded[1:, :-1]

    # iterate over elements in the sublattice
    for row in prange(x0.shape[0]):
        for col in prange(x0.shape[1]):

            d_p1, d_m1 = 0., 0.

            # horizontal
            x0_sub = d0_h[row+3, col:col+6]
            d_m1 += GetLocalDistortion(x0_sub, x0_sub-response_p1, sigma, gamma)
            d_p1 += GetLocalDistortion(x0_sub, x0_sub+response_p1, sigma, gamma)

            # vertical
            x0_sub = d0_v[row:row+6, col+3]
            d_m1 += GetLocalDistortion(x0_sub, x0_sub-response_p1, sigma, gamma)
            d_p1 += GetLocalDistortion(x0_sub, x0_sub+response_p1, sigma, gamma)

            # diagonal
            x0_sub = np.array([d0_d[row+i, col+i] for i in range(6)])
            d_m1 += GetLocalDistortion(x0_sub, x0_sub-response_p1, sigma, gamma)
            d_p1 += GetLocalDistortion(x0_sub, x0_sub+response_p1, sigma, gamma)

            # minor diagonal
            x0_sub = np.array([d0_m[row+i, col+5-i] for i in range(6)])
            d_m1 += GetLocalDistortion(x0_sub, x0_sub-response_p1, sigma, gamma)
            d_p1 += GetLocalDistortion(x0_sub, x0_sub+response_p1, sigma, gamma)

            #
            rho_p1[row, col] = d_p1
            rho_m1[row, col] = d_m1

    #
    return rho_p1, rho_m1


def compute_cost_adjusted(
    x0: np.ndarray,
    *,
    sigma: float = 1,
    gamma: float = 1,
    wet_cost: float = 1e8,
):
    """Computes HUGO cost with wet-cost adjustments.

    :param x0: uncompressed (pixel) cover image
        of shape [height, width]
    :type x0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param sigma: additive parameter
    :type sigma: float
    :param gamma: exponent parameter
    :type gamma: float
    :param wet_cost: wet cost for unembeddable coefficients
    :type wet_cost: float
    :return: costs for +1 and -1 changes
    :rtype: tuple of `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> # TODO
    """
    assert len(x0.shape) == 2, 'single channel expected'

    # cover to double
    x0 = x0.astype('double')

    # create mirror padded cover image
    x0_padded = np.pad(x0, [[3]*2, [3]*2], 'symmetric')

    # Compute costmap
    rho_p1, rho_m1 = compute_cost(
        x0=x0,
        x0_padded=x0_padded,
        sigma=sigma,
        gamma=gamma,
    )

    # Assign wet cost
    rho_p1[np.isinf(rho_p1) | np.isnan(rho_p1) | (rho_p1 > wet_cost)] = wet_cost
    rho_m1[np.isinf(rho_m1) | np.isnan(rho_m1) | (rho_m1 > wet_cost)] = wet_cost

    # Do not embed +-1 if the pixel has min/max value
    rho_p1[x0 >= 255] = wet_cost
    rho_m1[x0 <= 0] = wet_cost

    return rho_p1, rho_m1



"""Implementation of EBS cost.

The method was introduced in
C. Wang, et al. An efficient JPEG steganographic scheme based on block entropy of DCT coefficients.
IEEE ICASSP, 2012.

It was used in the ALASKA challenge.

Author: Martin Benes
Affiliation: University of Innsbruck
"""  # noqa: E501

import enum
import numpy as np


class Implementation(enum.Enum):
    """EBS implementation to choose from."""

    EBS_ORIGINAL = enum.auto()
    """Original EBS implementation by Remi Cogranne."""
    EBS_FIX_WET = enum.auto()
    """EBS implementation with wet-cost fixes in not-SI case."""


def _block_entropy(b: np.ndarray) -> float:
    """Computes block entropy of flattened block b.

    :param b:
    :type b: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :return:
    :rtype: float
    """
    b = b[1:]  # remove DC
    if (b == 0).all():
        return 0.
    values, counts = np.unique(b[b != 0], return_counts=True)
    Ps = counts.astype('float64') / np.sum(b != 0)
    H = -np.sum(Ps * np.log(Ps))  # mismatch with Matlab
    return H


def compute_cost(
    y0: np.ndarray,
    qt: np.ndarray,
    *,
    rounding_error: np.ndarray = None,
    theta: float = 2,
) -> np.ndarray:
    """Compute EBS cost.

    Near-equivalent to Remi Cogranne's implementation.
    The difference comes from numerical difference in log/np.log.

    :param y0:
    :type y0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param qt:
    :type y0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param rounding_error:
    :type rounding_error:
    :param theta: distortion parameter, 2 by default (from paper)
    :type theta: float
    :return:
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> # TODO
    """
    # block entropy cost
    block_entropies = np.apply_along_axis(
        _block_entropy,
        axis=1,
        arr=y0.reshape(-1, 64),
    ).reshape(*y0.shape[:2], 1, 1)

    # block distortion
    block_distortion = 1/(block_entropies**theta)
    block_distortion = np.repeat(np.repeat(block_distortion, 8, 2), 8, 3)

    # flipping distortion
    if rounding_error is not None:
        # compute quantized DCT xbar_i = round(x_qi)
        # compute rounding error e_ri = xbar_i - x_qi
        # compute quantization error e_qi = xbar_i*qi - xi
        # compute flipping cost rho_flip_i = [(np.abs(y' - x_qi)-.5)*qi]**2
        raise NotImplementedError('distortion from precover not implemented')
    else:
        qti = qt.copy()
        for i in range(2):
            qti = np.expand_dims(qti, i)
            qti = np.repeat(qti, y0.shape[i], i)
        flip_distortion = qti**2

    # final cost
    return block_distortion * flip_distortion


def compute_cost_adjusted(
    y0: np.ndarray,
    qt: np.ndarray,
    *,
    x0: np.ndarray = None,
    theta: float = 2,
    implementation: Implementation = Implementation.EBS_FIX_WET,
    wet_cost: float = 10**13,
) -> np.ndarray:
    """Computes the costmap and prepares the costmap for ternary embedding.

    :param y0: quantized cover DCT coefficients
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :type y0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param qt: quantization table
        of shape [8, 8]
    :type qt: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param x0: precover for rounding error computation
        of shape [num_vertical_blocks*8, num_horizontal_blocks*8]
    :param theta: cost parameter, 2 by default (from paper)
    :type theta: float
    :param implementation: choose EBS implementation
    :type implementation: :class:`Implementation`
    :param wet_cost: cost for unembeddable coefficients
    :type wet_cost: float
    :return: probability maps for +1 and -1 changes
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :rtype: tuple of `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> (rho_p1, rho_m1) = cl.ebs.compute_cost_adjusted(
    ...   y0=im_dct.Y,  # DCT
    ...   qt=im_dct.qt[0])  # QT
    >>> im_dct.Y += cl.simulate.ternary(
    ...   rhos=(rho_p1, rho_m1)  # costs of +1 and -1 changes
    ...   alpha=0.4,  # alpha
    ...   n=im_dct.Y.size,  # cover size
    ...   seed=12345)  # seed
    """
    # Compute rounding error
    if x0 is not None:
        raise NotImplementedError('side-informed EBS not implemented')
    else:
        rounding_error = None

    # Compute cost
    rho = compute_cost(
        y0=y0,
        qt=qt,
        rounding_error=rounding_error,
        theta=theta,
    )

    # Adjust embedding costs
    rho = rho + 10**-4

    # Assign wet cost
    rho[np.isinf(rho) | np.isnan(rho) | (rho > wet_cost)] = wet_cost

    # Do not embed into following modes (from Remi's implementation)
    if implementation == Implementation.EBS_ORIGINAL:
        rho[:, :, 0, 0] = wet_cost
        rho[:, :, 0, 4] = wet_cost
        rho[:, :, 4, 0] = wet_cost
        rho[:, :, 4, 4] = wet_cost

    elif implementation == Implementation.EBS_FIX_WET:
        # Avoid 04 coefficients with e = 0.5
        if x0 is not None:
            max_cost_mat = np.zeros(shape=rho.shape, dtype=bool)
            max_cost_mat[:, :, 0, 0] = True
            max_cost_mat[:, :, 4, 0] = True
            max_cost_mat[:, :, 0, 4] = True
            max_cost_mat[:, :, 4, 4] = True
            rho[max_cost_mat & (np.abs(rounding_error) > 0.4999)] = wet_cost
    else:
        raise NotImplementedError(f'unknown implementation {implementation}')

    # # Do not embed +-1 if the DCT coefficient has extreme value
    # rho[cover_dct_coeffs >= 1023] = wet_cost
    # rho[cover_dct_coeffs <= -1023] = wet_cost
    # return rho

    # Do not embed +1 if the DCT coefficient has max value
    rho_p1 = np.copy(rho)
    rho_p1[y0 >= 1023] = wet_cost

    # Do not embed -1 if the DCT coefficient has min value
    rho_m1 = np.copy(rho)
    rho_m1[y0 <= -1023] = wet_cost

    return rho_p1, rho_m1



"""

Author: Martin Benes
Affiliation: University of Innsbruck
"""

import numpy as np

from .. import nsF5


def simulate_single_channel(*args, **kw) -> np.ndarray:
    """Simulates F5 embedding at an embedding rate at an embedding rate into single-channel cover and returns stego.

    This is done by simulating nsF5 first, and then re-introducing the shrinkage.

    :return: quantized stego DCT coefficients
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> im_dct.Y = cl.F5.simulate_single_channel(
    ...   y0=im_dct.Y,  # DCT
    ...   alpha=0.4,  # alpha
    ...   seed=12345)  # seed
    """
    return nsF5.simulate_single_channel(*args, add_shrinkage=True, **kw)



"""
J-MiPOD implementation based on https://codeocean.com/capsule/7800700/tree/v4
Author: Martin Benes, Benedikt Lorch
Affiliation: University of Innsbruck

This Python implementation is ported from the original Matlab implementation provided by the paper authors. Please find the license of the original implementation below.
-------------------------------------------------------------------------
Copyright (c) 2020 Remi Cogranne, UTT (Troyes University of Technology). All Rights Reserved.
-------------------------------------------------------------------------
This code is provided by the author under Creative Common License (CC BY-NC-SA 4.0) which, as explained on this webpage https://creativecommons.org/licenses/by-nc-sa/4.0/ Allows modification, redistribution, provided that:
* You share your code under the same license;
* That you give credits to the authors;
* The code is used only for non-commercial purposes (which includes education and research)
-------------------------------------------------------------------------
The authors hereby grant the use of the present code without fee, and without a written agreement under compliance with aforementioned and provided and the present copyright notice appears in all copies. The program is supplied "as is," without any accompanying services from the UTT or the authors. The UTT does not warrant the operation of the program will be uninterrupted or error-free. The end-user understands that the program was developed for research purposes and is advised not to rely exclusively on the program for any reason. In no event shall the UTT or the authors be liable to any party for any consequential damages. The authors also fordid any practical use of this code for communication by hiding data into JPEG images.
-------------------------------------------------------------------------
"""

import os.path
import numpy as np
import scipy.io
import scipy.optimize
import scipy.signal
from typing import Tuple
import warnings

from .. import tools


def wiener2(
    im: np.ndarray,
    kernel_size: Tuple[int],
    *,
    noise: float = None,
) -> np.ndarray:
    """
    Adaptive noise removal.

    Uses zero padding to match the Matlab implementation.

    From the Matlab docstring:
    Wiener lowpass-filters an intensity image that has been degraded by constant power additive noise.
    This filter uses a pixel-wise adaptive Wiener method based on statistics estimated from a local neighborhood of each pixel.

    :param im: 2D ndarray
    :type im: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param kernel_size: 2-tuple containing the window size for estimating local mean and standard deviation
    :type kernel_size: tuple of int
    :param noise: power of assumed noise
    :type noise: float
    :return: filtered image
    :rtype:
    """

    # Parse given kernel size
    if kernel_size is None:
        # Default to isotropic kernel of size 3
        kernel_size = [3] * im.dim
    kernel_size = np.asarray(kernel_size)
    if kernel_size.shape == ():
        # Expand integer kernel size to the image dimensions
        kernel_size = np.repeat(kernel_size.item(), im.ndim)

    num_kernel_elements = np.prod(kernel_size)

    # Estimate the local mean.
    # The Matlab implementation uses zero-padding.
    local_mean = scipy.signal.correlate2d(im, np.ones(kernel_size), mode="same", boundary="fill") / num_kernel_elements

    # Estimate the local variance
    local_mean_squared = scipy.signal.correlate2d(im ** 2, np.ones(kernel_size), mode="same", boundary="fill") / num_kernel_elements
    local_variance = local_mean_squared - local_mean ** 2
    # FIX: for flat images
    # if np.isnan(local_variance).any():
    #     warnings.warn('invalid variance in flat areas, clipping')
    #     local_variance[np.isnan(local_variance)] = 0
    #     local_variance = np.clip(local_variance, a_min=0.01, a_max=None)
    if (local_variance == 0).any():
        warnings.warn('invalid variance in flat areas, clipping')
        local_variance = np.clip(local_variance, a_min=1e-8, a_max=None)

    # Estimate the noise power if needed
    if noise is None:
        noise = np.mean(local_variance)

    res = im - local_mean
    res *= (1 - noise / np.maximum(local_variance, noise))  # Avoid division by zero variance. If local variance < noise, we are going to replace it with the local mean later anyway.
    res += local_mean
    out = np.where(local_variance < noise, local_mean, res)

    # Sanity check
    out2 = local_mean + (np.maximum(0, local_variance - noise) / np.maximum(local_variance, noise)) * (im - local_mean)
    assert np.allclose(out, out2)

    return out


def im2col(
    img: np.ndarray,
    window_shape: Tuple[int],
    *,
    step_size: int = 1,
) -> np.ndarray:
    """
    Return sliding window over the given 2D image
    :param img: input image
    :type img: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param window_shape: 2-tuple with the window size
    :type window_shape:
    :param step_size: step size
    :type step_size: int
    :return: view of shape [window_height, window_width, num_windows]
    """
    # Determine number of output windows
    img_height, img_width = img.shape
    num_rows = img_height - window_shape[0] + 1
    num_cols = img_width - window_shape[1] + 1
    shape = (window_shape[0], window_shape[1], num_rows, num_cols)

    s0, s1 = img.strides
    strides = s0, s1, s0, s1

    out_view = np.lib.stride_tricks.as_strided(img, shape=shape, strides=strides)
    return out_view.reshape((window_shape[0] * window_shape[1], num_rows * num_cols))[:, ::step_size]


def pad_copy_borders(x, pad_size):
    pad_top, pad_bottom = pad_size[0]
    pad_left, pad_right = pad_size[1]

    # Zero-padding leaves the original array in the center.
    x_padded = np.pad(x, pad_width=pad_size, mode="constant")

    # Fill the top rows by copying from the next 8 rows
    if pad_top > 0:
        x_padded[:pad_top, :] = x_padded[pad_top:2 * pad_top, :]

    # Fill the bottom rows
    if pad_bottom > 0:
        x_padded[-pad_bottom:, :] = x_padded[-2 * pad_bottom:-pad_bottom, :]

    # Fill left
    if pad_left > 0:
        x_padded[:, :pad_left] = x_padded[:, pad_left:2 * pad_left]

    # Fill right
    if pad_right > 0:
        x_padded[:, -pad_right:] = x_padded[:, -2 * pad_right:-pad_right]

    return x_padded


def invxlnx3_fast(y, f, excess_values_num_iter=10):
    """
    Fast solver of y = x * log(x - 2) parallelized over all pixels
    :param y: pixels?
    :param f: lookup table for solving y = x * log(x - 2) for x, with resolution 100
    :param excess_values_num_iter: number of iterations for computing the values that exceed the precomputed range
    :return: x
    """

    # The lookup table only contains values up to y = 1000.
    i_large = y > 1000
    i_small = y <= 1000

    # The lookup table uses a finer resolution with values y = [0, 1 / resolution, 2 / resolution, ..., 1000].
    # The default resolution is 100.
    iyL = np.floor(y[i_small] / 0.01).astype(int)
    iyR = iyL + 1

    # Clamp values that would exceed the lookup table
    iyR[iyR > 100000] = 100000

    x = np.zeros_like(y)
    x[i_small] = f[iyL] + (y[i_small] - iyL * 0.01) * (f[iyR] - f[iyL])

    # For values outside the precomputed range, we iteratively compute the result.
    z = y[i_large] / np.log(y[i_large] - 2)
    for _ in range(excess_values_num_iter):
        z = y[i_large] / np.log(z - 2)

    x[i_large] = z

    return x


def ternary_entropy(probs):
    return tools.entropy(probs, probs, base=np.e)

# def ternary_entropy(probs):
#     p0 = 1 - 2 * probs
#     P = np.hstack([p0, probs, probs])
#     H = -P * np.log(P)
#     #
#     eps = np.finfo(np.float32).eps
#     H[P < eps] = 0
#     #
#     return np.nansum(H)


def ternary_probs(
    fisher_information: float,
    payload: int,
    max_num_iterations: int = 30,
    excess_values_num_iter: int = 10,
) -> float:
    """
    Determine the change rates for each cover element by searching for the Lagrangian multiplier lambda that optimizes the payload constraint.
    :param fisher_information: fisher information
    :type fisher_information: float
    :param payload: target payload
    :type payload: int
    :param max_num_iterations: maximum number of binary search iterations
    :type max_num_iterations: int
    :param excess_values_num_iter: number of iterations for computing the values that exceed the precomputed range
    :type excess_values_num_iter: int
    :return: change rate per cover element
    :rtype: float
    """
    ixlnx3 = load_lookup_table()

    fisher_information_flat = fisher_information.flatten(order="F")

    # Initial search interval for lambda
    L = 10 ** 3
    R = 10 ** 6

    # fL, fR > 0 when they fit the desired payload
    # Note that smaller lambdas lead to higher ternary entropy.
    fL = ternary_entropy(1. / invxlnx3_fast(L * fisher_information_flat, ixlnx3, excess_values_num_iter=excess_values_num_iter)) - payload
    fR = ternary_entropy(1. / invxlnx3_fast(R * fisher_information_flat, ixlnx3, excess_values_num_iter=excess_values_num_iter)) - payload

    # If the range [L, R] does not cover the desired payload, enlarge the search interval.
    # fL < 0 and fR < 0: smallest lambda is still too big, expand search interval to the left by halving L
    # fL > 0 and fR > 0: highest lambda is still too low, expand search interval to the right by doubling R
    while fL * fR > 0:
        if fL > 0:
            R = 2 * R
            fR = ternary_entropy(1. / invxlnx3_fast(R * fisher_information_flat, ixlnx3, excess_values_num_iter=excess_values_num_iter)) - payload
        else:
            L = L / 2
            fL = ternary_entropy(1. / invxlnx3_fast(L * fisher_information_flat, ixlnx3, excess_values_num_iter=excess_values_num_iter)) - payload

    # At this point, L is below and R is above the desired lambda. Hence, fL > 0 and fR < 0.

    # Search for lambda (variable M) in the specified interval
    i = 0

    # fM is the distance between the target payload and the payload with the current lambda
    fM = 1

    # Record lambda and the distance between the target payload and the current payload
    TM = np.zeros((max_num_iterations, 2))

    # Binary search for lambda
    # Stopping criteria:
    # (1) The distance between the target payload and the current payload is less than or equal to 0.0001.
    # (2) Stop after a maximum of 30 iterations.
    while np.abs(fM) > 0.0001 and i < max_num_iterations:
        # Mid of the interval
        M = (L + R) / 2

        # Evaluate distance between target payload and current payload
        fM = ternary_entropy(1. / invxlnx3_fast(M * fisher_information_flat, ixlnx3, excess_values_num_iter=excess_values_num_iter)) - payload

        # fL is > 0 by construction
        if fL * fM < 0:
            # If the current lambda fits the payload, then we can cap the search interval at the top.
            R = M
            fR = fM

        else:
            # fM > 0: We can cap the search interval at the bottom.
            L = M
            fL = fM

        # Store current results
        TM[i, 0] = fM
        TM[i, 1] = M

        # Increment iteration counter
        i += 1

    if max_num_iterations == i:
        # Maximum number of iterations reached

        # Find the lambda that achieved the lowest absolute distance to the target payload
        M = TM[np.argmin(np.abs(TM[:, 0])), 1]

    # Compute beta using the lambda found
    # beta is the change rate
    beta = 1. / invxlnx3_fast(M * fisher_information_flat, ixlnx3, excess_values_num_iter=excess_values_num_iter)

    return beta


def prepare_lookup_table(max_val=1000, resolution=100, ftol=1e-3):
    """
    Prepare a lookup table for inverting the function y = x * log(x - 2).

    The values in the lookup table correspond to y in [0, max_value].
    The lookup table entries resolve with finer resolution, i.e., [0, 1 / resolution, 2 / resolution, ..., max_value].

    If you want to know which x gives the value y, do

    x = lut[ floor(y * resolution) ]
    x * log(x - 2) should be close to y

    :param max_val: highest y value in the lookup table (inclusive)
    :param resolution: precision of the lookup table, should be >= 1.
    :param ftol: tolerance in function value; stopping criterion of the minimum search
    :return: lookup table of the desired size
    """

    def fun(x):
        return x * np.log(x - 2)

    num_entries = max_val * resolution + 1
    xs = np.zeros(num_entries)

    for i in range(num_entries):
        y = i / resolution

        # We want to minimize the absolute distance between y and f(x)
        def find_inverse(x):
            return np.abs(y - fun(x))

        x = scipy.optimize.fmin(func=find_inverse, x0=3, xtol=1e-6, ftol=ftol, disp=0)[0]

        # Quick sanity check
        assert np.isclose(fun(x), y, atol=ftol)

        # Store the result
        xs[i] = x

    return xs


def load_lookup_table():
    mat_file = os.path.join(os.path.dirname(__file__), "ixlnx3.mat")
    npy_file = os.path.join(os.path.dirname(__file__), "ixlnx3.npz")

    if os.path.exists(mat_file):
        # Loading lookup table from Matlab
        ixlnx3 = scipy.io.loadmat(mat_file)["ixlnx3"].flatten()

    elif os.path.exists(npy_file):
        # Loading our own pre-computed lookup table
        ixlnx3 = np.load(npy_file)["ixlnx3"]

    else:
        # No lookup table yet
        print("Lookup table does not exist yet. Preparing and storing lookup table. This can take a few minutes.")

        # Need to create the lookup table
        ixlnx3 = prepare_lookup_table(max_val=1000, resolution=100)

        # Store lookup table
        np.savez(npy_file, ixlnx3=ixlnx3)

    return ixlnx3



"""

Author: Martin Benes, Benedikt Lorch
Affiliation: University of Innsbruck
"""

# simulate
from ._simulate import simulate
from ._costmap import compute_cost_adjusted, Change, Location

#
from . import _simulate
from . import _costmap

__all__ = [
    '_simulate',
    '_costmap',
    'compute_cost_adjusted',
    'simulate_single_channel',
    'Change',
    'Location',
]



"""

Author: Martin Benes, Benedikt Lorch
Affiliation: University of Innsbruck
"""

import numpy as np

from ._costmap import compute_cost_adjusted, Implementation
from .. import simulate
from .. import tools


def simulate_single_channel(
    y0: np.ndarray,
    qt: np.ndarray,
    alpha: float,
    *,
    wet_cost: float = 10**13,
    implementation: Implementation = Implementation.EBS_ORIGINAL,
    generator: str = 'MT19937',
    order: str = 'C',
    seed: int = None,
) -> np.ndarray:
    """Simulates EBS embedding at an embedding rate into single-channel cover,
    and returns stego.

    EBS was introduced in
    C. Wang, et al. An efficient JPEG steganographic scheme based on block entropy of DCT coefficients.
    IEEE ICASSP, 2012.

    The details of the methods are described in the
    `glossary <https://conseal.readthedocs.io/en/latest/glossary.html#entropy-block-stego>`__.

    :param y0: quantized cover DCT coefficients
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :type y0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param qt: quantization table
        of shape [8, 8]
    :type qt: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param alpha: embedding rate
        in bits per nzAC coefficient
    :type alpha: float
    :param wet_cost: wet cost for unembeddable coefficients
    :type wet_cost: float
    :param implementation: choose EBS implementation
    :type implementation: :class:`Implementation`
    :param generator: type of PRNG used by embedding simulator
    :type generator: `numpy.random.Generator <https://numpy.org/doc/stable/reference/random/generator.html>`__
    :param seed: random seed for embedding simulator
    :type seed: int
    :return: quantized stego DCT coefficients
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> im_dct.Y = cl.ebs.simulate_single_channel(
    ...   y0=im_dct.Y,      # DCT
    ...   qt=im_dct.qt[0],  # QT
    ...   alpha=0.4,        # alpha
    ...   seed=12345)       # seed
    """
    # Count number of embeddable DCT coefficients
    N, M, _, _ = y0.shape
    if implementation == Implementation.EBS_ORIGINAL:
        num_DCT_coeffs = y0.size - 4*N*M
    elif implementation == Implementation.EBS_FIX_WET:
        num_DCT_coeffs = y0.size
    else:
        raise NotImplementedError(f'unknown implementation {implementation}')

    # Compute cost for embedding into the quantized DCT coefficients
    # of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    rho_p1, rho_m1 = compute_cost_adjusted(
        y0=y0,
        qt=qt,
        implementation=implementation,
        wet_cost=wet_cost,
    )

    # STC simulation
    rho_p1 = tools.dct.jpeglib_to_jpegio(rho_p1)
    rho_m1 = tools.dct.jpeglib_to_jpegio(rho_m1)
    delta = simulate.ternary(
        rho_p1=rho_p1,
        rho_m1=rho_m1,
        alpha=alpha,
        n=num_DCT_coeffs,
        generator=generator,
        order=order,
        seed=seed,
    )
    delta = tools.dct.jpegio_to_jpeglib(delta)

    #
    return y0 + delta



"""

Author: Benedikt Lorch, Martin Benes
Affiliation: University of Innsbruck
"""

# costmap
from ._costmap import compute_cost_adjusted, evaluate_cost, Implementation

# simulate
from . import _simulate
from ._simulate import simulate_single_channel

# abbreviations of enum
JUNIWARD_ORIGINAL = Implementation.JUNIWARD_ORIGINAL
JUNIWARD_FIX_OFF_BY_ONE = Implementation.JUNIWARD_FIX_OFF_BY_ONE

__all__ = [
    '_costmap',
    '_simulate',
    'compute_cost_adjusted',
    'evaluate_cost',
    'simulate_single_channel',
    'Implementation',
    'JUNIWARD_ORIGINAL',
    'JUNIWARD_FIX_OFF_BY_ONE',
]



"""

Author: Martin Benes
Affiliation: University of Innsbruck
"""

import numpy as np
from scipy.stats import rankdata


def probability(
    rhos: np.ndarray,
    alpha: float,
    e: float = 2
) -> np.ndarray:
    """Returns the probabilities for a selection channel steganography.

    A quick and dirty implementation. Will be updated in the future.
    See the code comments for more information.

    :param rhos: embedding costs
    :type rhos: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param alpha: embedding rate
    :type alpha: float
    :param e: embedding efficiency
    :type e: float
    :return: change probabilities
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    """
    # cost rank
    ranks = rankdata(rhos.reshape(-1), method='ordinal') - 1
    """smallest cost is 0, largest cost is N-1"""
    # probability of this cost being lower (out of 2)
    ps = [
        (rhos.size - rank) / (rhos.size - 1)
        for rank in ranks
    ]
    """normalize the ranks to [0,1]"""
    # probability of this cost being lower from k
    inv_alpha = np.floor(1 / alpha)  # block size
    ps = np.array([
        p**(inv_alpha - 1) / e
        for p in ps
    ]).reshape(rhos.shape)
    """extends from 2 to k.

    For e=2, this is

    k  p
    1  p^0/2=0.5 (behaves wierd for alpha>0.5)
    2  p^1/2
    3  p^2/2
    4  p^3/2

    Quick and dirty implementations. Will be updated in the future.

    Know limitations (for the moment):
    - assumes sampling with replacement (should be ok for large covers)
    """

    #
    return ps.reshape(rhos.shape)



"""Simulator of binary embedding.

Author: Martin Benes, Benedikt Lorch
Affiliation: University of Innsbruck

This implementation builds on the Matlab implementation provided by the DDE lab. Please find that license of the original implementation below.
-------------------------------------------------------------------------
Copyright (c) 2013 DDE Lab, Binghamton University, NY. All Rights Reserved.
Permission to use, copy, modify, and distribute this software for educational, research and non-profit purposes, without fee, and without a written agreement is hereby granted, provided that this copyright notice appears in all copies. The program is supplied "as is," without any accompanying services from DDE Lab. DDE Lab does not warrant the operation of the program will be uninterrupted or error-free. The end-user understands that the program was developed for research purposes and is advised not to rely exclusively on the program for any reason. In no event shall Binghamton University or DDE Lab be liable to any party for direct, indirect, special, incidental, or consequential damages, including lost profits, arising out of the use of this software. DDE Lab disclaims any warranties, and has no obligations to provide maintenance, support, updates, enhancements or modifications.
-------------------------------------------------------------------------
"""  # noqa: E501

import numpy as np
from typing import Tuple, Callable

from ._optim import get_objective, calc_lambda, Sender


def probability(
    rhos: Tuple[np.ndarray],
    alpha: float,
    n: int,
    *,
    e: float = None,
    objective: Callable = None,
    sender: Sender = Sender.PAYLOAD_LIMITED_SENDER,
) -> Tuple[np.ndarray, float]:
    """Convert binary distortion to binary probability.

    :param rhos: distortion tensor for +-1 change
    :type rhos: tuple of `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param alpha: embedding rate in bits per element, generally a constraint for the specified sender type
    :type alpha: float
    :param n: cover size/number of cover elements
    :type n: int
    :param e: Embedding efficiency,
        in bits per element.
        If not provided, optimal coding is assumed.
    :type e: float
    :param objective: Objective function to maximize.
        Entropy by default
    :type objective: callable
    :param sender: type of sender, changes the semantics of parameter alpha,
        PLS by default
    :return: tuple ((p_pm1), lmbda), where
        p_pm1 is the probability of +-1 change, and
        lbda is the determined lambda.
    :rtype: tuple

    :Example:

    >>> (p_pm1,), lbda = cl.simulate._binary.probability(
    ...   rhos=(rho_pm1,),  # costs for +1 and -1
    ...   alpha=0.4,        # embedding rate, 0.4 message bits per element
    ...   n=im_dct.Y.size)  # cover size
    >>> im_dct.Y += cl.simulate._binary.simulate(
    ...   ps=(p_pm1,),  # probability of +1 and -1
    ...   seed=12345)   # seed
    """
    if objective is None:
        objective = get_objective(e=e, q=2, sender=sender)

    message_length = int(np.round(alpha * n))
    lbda = calc_lambda(rhos, message_length, n, objective)
    #
    ps, _ = objective(lbda=lbda, rhos=rhos)
    return ps, lbda


def simulate(
    ps: Tuple[np.ndarray],
    generator: str = None,
    order: str = 'C',
    seed: int = None,
) -> np.ndarray:
    """Simulates changes using the given probability maps.

    :param ps: probability tensor for +-1 change,
        of an arbitrary shape
    :type p_p1: tuple of `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param generator: random number generator to choose,
        None (numpy default) or 'MT19937' (used by Matlab)
    :type generator: str
    :param order: order of changes,
        'C' (C-order, column-row) or 'F' (F-order, row-column).
    :type order: str
    :param seed: random seed for embedding simulator
    :type seed: int
    :return: Simulated ternary changes in the cover, 0 (keep), +1 or -1.
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> (p_pm1,), lbda = cl.simulate._binary.probability(
    ...   rhos=(rho_pm1,),  # costs for +1 and -1
    ...   alpha=0.4,        # embedding rate, 0.4 message bits per element
    ...   n=im_dct.Y.size)  # cover size
    >>> im_dct.Y += cl.simulate._binary.simulate(
    ...   ps=(p_pm1,),  # probability of +1 and -1
    ...   seed=12345)   # seed
    """
    assert len(ps) in {1, 2}
    (p_pm1,) = ps[:1]

    # Select random number generator
    if generator is None:  # numpy default generator
        rng = np.random.default_rng(seed=seed)
        rand_change = rng.random(p_pm1.shape)

    elif generator == 'MT19937':  # Matlab generator
        prng = np.random.RandomState(seed)
        rand_change = prng.random_sample(p_pm1.shape)

    else:
        try:
            rand_change = generator(p_pm1.shape)
        except Exception:
            raise NotImplementedError(f'unsupported generator {generator}')

    # Order of changes
    if order is None or order == 'C':
        pass

    elif order == 'F':
        rand_change = rand_change.reshape(-1).reshape(p_pm1.shape, order='F')

    else:
        raise NotImplementedError(f'Given order {order} is not implemented')

    # Set up ndarray with simulated changes
    delta = np.zeros(p_pm1.shape, dtype='int8')

    delta[rand_change <= p_pm1] = 1
    return delta


def binary(
    rhos: Tuple[np.ndarray],
    alpha: float,
    n: int,
    *,
    e: float = None,
    objective: Callable = None,
    sender: Sender = Sender.PAYLOAD_LIMITED_SENDER,
    **kw,
) -> np.ndarray:
    """Simulates binary embedding given distortion and embedding rate.

    :param rhos: cost for +-1 changes,
        of an arbitrary shape
    :type rhos: tuple of `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param alpha: embedding rate,
        in bits per element
    :type alpha: float
    :param n: cover size/number of cover elements
    :type n: int
    :param objective: Objective function to maximize.
        Entropy by default
    :type objective: callable
    :return: Simulated difference image to be added to the cover, 0 (keep), 1 or -1 (change).
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> (rho_pm1,) = cl.uerd.compute_cost_adjusted(
    ...   cover_dct_coeffs=im_dct.Y,  # DCT
    ...   quantization_table=im_dct.qt[0])  # QT
    >>> im_dct.Y += cl.simulate.binary(
    ...   rhos=(rho_pm1,),  # cost of +-1 change
    ...   alpha=0.4,  # alpha
    ...   n=im_dct.Y.size,  # cover size
    ...   seed=12345)  # seed
    """
    ps, lbda = probability(
        rhos=rhos,
        alpha=alpha,
        n=n,
        e=e,
        objective=objective,
        sender=sender,
    )
    return simulate(ps, **kw)



"""

Author: Martin Benes, Benedikt Lorch
Affiliation: University of Innsbruck
"""

#
from ._simulate import simulate_single_channel

#
from . import _simulate

__all__ = [
    '_simulate',
    'simulate_single_channel',
]



"""Simulation of the UERD steganography method.

We thank Patrick Bas for sharing his implementation of UERD with us.

Author: Martin Benes, Benedikt Lorch
Affiliation: University of Innsbruck
"""

import numpy as np

from ._costmap import compute_cost_adjusted
from ..simulate import _ternary
from .. import tools


def simulate_single_channel(
    y0: np.ndarray,
    qt: np.ndarray,
    alpha: float,
    *,
    payload_mode: str = "bpnzAC",
    wet_cost: float = 10**13,
    seed: int = None,
) -> np.ndarray:
    """Simulates UERD embedding at an embedding rate into single-channel cover,
    and returns stego.

    UERD was introduced in
    L. Guo, et al. Using Statistical Image Model for JPEG Steganography: Uniform Embedding Revisited.
    IEEE TIFS, 2015.

    The details of the methods are described in the
    `glossary <https://conseal.readthedocs.io/en/latest/glossary.html#uniform-embedding-revisited-distortion-uerd>`__.

    :param y0: quantized DCT coefficients
        of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :type y0: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param qt: quantized table of shape [8, 8]
    :type qt: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__
    :param alpha: embedding rate, in unit specified by payload_mode
    :type alpha: float
    :param payload_mode: unit used by embedding rate, either
        "bpc" (bits per DCT coefficient), or
        "bpnzAC" (bits per non-zero DCT AC coefficient).
    :type payload_mode: str
    :param wet_cost: wet cost for unembeddable coefficients
    :type wet_cost: float
    :param seed: random seed for embedding simulator
    :type seed: int
    :return: stego DCT coefficients of shape [num_vertical_blocks, num_horizontal_blocks, 8, 8]
    :rtype: `np.ndarray <https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html>`__

    :Example:

    >>> im_dct.Y = cl.uerd.simulate_single_channel(
    ...   y0=im_dct.Y,  # DCT
    ...   qt=im_dct.qt[0],  # QT
    ...   alha=0.4,  # alpha
    ...   seed=12345)  # seed
    """
    if np.isclose(alpha, 0):
        return y0

    # Compute distortion
    rhos = compute_cost_adjusted(
        y0=y0,
        qt=qt,
        wet_cost=wet_cost,
    )

    # Determine number of available coefficients
    if "bpc" == payload_mode:
        n = y0.size
    elif "bpnzAC" == payload_mode:
        n = tools.dct.nzAC(y0)
    else:
        raise ValueError("Unknown payload mode")

    # Catch no nzAC (fails otherwise)
    if n == 0:
        raise ValueError('There are no non-zero AC coefficients for embedding')

    # simulator
    ps, lbda = _ternary.probability(
        rhos=rhos,
        alpha=alpha,
        n=n,
    )

    delta = _ternary.simulate(
        ps=ps,
        seed=seed,
    )

    return y0 + delta
